"""Generate VTIMEZONE components from actual timezone information.

When we generate VTIMEZONE from actual tzinfo instances of
- dateutil
- zoneinfo
- pytz

Then, we cannot assume that the future information stays the same but
we should be able to create tests that work for the past.

See https://github.com/collective/icalendar/issues/722
"""

from datetime import date, datetime, timedelta
from re import findall

import pytest
from dateutil.tz import gettz

from icalendar import Calendar, Component, Event, Timezone
from icalendar.compatibility import zoneinfo
from icalendar.timezone import tzid_from_tzinfo, tzids_from_tzinfo

tzids = pytest.mark.parametrize(
    "tzid",
    [
        "Europe/Berlin",
        "Asia/Singapore",
        "America/New_York",
    ],
)


def assert_components_equal(c1: Component, c2: Component):
    """Print the diff of two components."""
    ML = 32
    ll1 = c1.to_ical().decode().splitlines()
    ll2 = c2.to_ical().decode().splitlines()
    pad = max(len(l) for l in ll1 if len(l) <= ML)
    diff = 0
    for l1, l2 in zip(ll1, ll2):
        a = len(l1) > 32 or len(l2) > 32
        print(
            a * "  " + l1,
            " " * (pad - len(l1)),
            a * "\n->" + l2,
            " " * (pad - len(l2)),
            "\tdiff!" if l1 != l2 else "",
        )
        diff += l1 != l2
    assert not diff, f"{diff} lines differ"


@tzids
def test_conversion_converges(tzp, tzid):
    """tzinfo -> VTIMEZONE -> tzinfo -> VTIMEZONE

    We can assume that both generated VTIMEZONEs are equivalent.
    """
    if tzp.uses_pytz():
        pytest.skip(
            "pytz will not converge on the first run. This is problematic. PYTZ-TODO"
        )
    tzinfo1 = tzp.timezone(tzid)
    assert tzinfo1 is not None
    generated1 = Timezone.from_tzinfo(tzinfo1)
    generated1["TZID"] = (
        "test-generated"  # change the TZID so we do not use an existing one
    )
    tzinfo2 = generated1.to_tz()
    generated2 = Timezone.from_tzinfo(tzinfo2, "test-generated")
    tzinfo3 = generated2.to_tz()
    generated3 = Timezone.from_tzinfo(tzinfo3, "test-generated")
    # pprint(generated1.get_transitions())
    # pprint(generated2.get_transitions())
    assert_components_equal(generated1, generated2)
    assert_components_equal(generated2, generated3)
    assert 2 <= len(generated1.standard + generated1.daylight) <= 3
    assert 2 <= len(generated2.standard + generated2.daylight) <= 3
    assert dict(generated1) == dict(generated2)
    assert generated1.to_ical().decode() == generated2.to_ical().decode()
    assert generated1.daylight == generated2.daylight
    assert generated1.standard == generated2.standard
    assert generated1 == generated2


@tzids
def both_tzps_generate_the_same_info(tzid, tzp):
    """We want to make sure that we get the same info for all timezone implementations.

    We assume that
    - the timezone implementations have the same info within the days we test
    - the timezone transitions times do not change because they are before last_date
    """
    # default generation
    tz1 = Timezone.from_tzid(tzid, tzp, last_date=date(2024, 1, 1))
    tzp.use_zoneinfo()  # we compare to zoneinfo
    tz2 = Timezone.from_tzid(tzid, tzp, last_date=date(2024, 1, 1))
    assert_components_equal(tz1, tz2)
    assert tz1 == tz2


@tzids
def test_tzid_matches(tzid, tzp):
    """Check the TZID."""
    tz = Timezone.from_tzinfo(tzp.timezone(tzid))
    assert tz["TZID"] == tzid


def test_do_not_convert_utc(tzp):
    """We do not need to convert UTC but it should work."""
    utc = Timezone.from_tzid("UTC")
    assert utc.daylight == []
    assert len(utc.standard) == 1
    standard = utc.standard[0]
    assert standard["TZOFFSETFROM"].td == timedelta(0)
    assert standard["TZOFFSETTO"].td == timedelta(0)
    assert standard["TZNAME"] == "UTC"


def test_berlin_time(tzp):
    """Test the Europe/Berlin timezone conversion."""
    tz = Timezone.from_tzid("Europe/Berlin")
    # we should have two timezones in it
    for x in tz.standard:
        print(x.name, x["TZNAME"], x["TZOFFSETFROM"].td, x["TZOFFSETTO"].td)
        print(x.to_ical().decode())
    assert len(tz.daylight) == 1
    assert len(tz.standard) in (1, 2), "We start in winter"
    dst = tz.daylight[-1]
    sta = tz.standard[-1]
    assert dst["TZNAME"] == "CEST"  # summer
    assert sta["TZNAME"] == "CET"
    assert dst["TZOFFSETFROM"].td == timedelta(hours=1)  # summer
    assert sta["TZOFFSETFROM"].td == timedelta(hours=2)
    assert dst["TZOFFSETTO"].td == timedelta(hours=2)  # summer
    assert sta["TZOFFSETTO"].td == timedelta(hours=1)


def test_range_is_not_crossed():
    first_date = datetime(2023, 1, 1)
    last_date = datetime(2024, 1, 1)

    def check(dt):
        assert first_date <= dt <= last_date

    tz = Timezone.from_tzid("Europe/Berlin", last_date=last_date, first_date=first_date)
    for sub in tz.standard + tz.daylight:
        check(sub.DTSTART)
        for rdate in sub.get("RDATE", []):
            check(rdate)


@tzids
def test_use_the_original_timezone(tzid, tzp):
    """When we get the timezone again, usually, we should use the
    one of the library/tzp."""
    tzinfo1 = tzp.timezone(tzid)
    assert tzinfo1 is not None
    generated1 = Timezone.from_tzinfo(tzinfo1)
    tzinfo2 = generated1.to_tz()
    assert type(tzinfo1) == type(tzinfo2)
    assert tzinfo1 == tzinfo2


@pytest.mark.parametrize(
    ("tzid", "dt", "tzname"),
    [
        ("Asia/Singapore", datetime(1970, 1, 1), "+0730"),
        ("Asia/Singapore", datetime(1981, 12, 31), "+0730"),
        ("Asia/Singapore", datetime(1981, 12, 31, 23, 10), "+0730"),
        ("Asia/Singapore", datetime(1981, 12, 31, 23, 34), "+0730"),
        ("Asia/Singapore", datetime(1981, 12, 31, 23, 59, 59), "+0730"),
        ("Asia/Singapore", datetime(1982, 1, 1), "+08"),
        ("Asia/Singapore", datetime(1982, 1, 1, 0, 1), "+08"),
        ("Asia/Singapore", datetime(1982, 1, 1, 0, 34), "+08"),
        ("Asia/Singapore", datetime(1982, 1, 1, 1, 0), "+08"),
        ("Asia/Singapore", datetime(1982, 1, 1, 1, 1), "+08"),
        ("Europe/Berlin", datetime(1970, 1, 1), "CET"),
        ("Europe/Berlin", datetime(2024, 3, 31, 0, 0), "CET"),
        ("Europe/Berlin", datetime(2024, 3, 31, 1, 0), "CET"),
        ("Europe/Berlin", datetime(2024, 3, 31, 2, 0), "CET"),
        ("Europe/Berlin", datetime(2024, 3, 31, 2, 59, 59), "CET"),
        ("Europe/Berlin", datetime(2024, 3, 31, 3, 0), "CEST"),
        ("Europe/Berlin", datetime(2024, 3, 31, 3, 0, 1), "CEST"),
        ("Europe/Berlin", datetime(2024, 3, 31, 4, 0), "CEST"),
        ("Europe/Berlin", datetime(2024, 10, 27, 0, 0), "CEST"),
        ("Europe/Berlin", datetime(2024, 10, 27, 1, 0), "CEST"),
        ("Europe/Berlin", datetime(2024, 10, 27, 2, 0), "CEST"),
        ("Europe/Berlin", datetime(2024, 10, 27, 2, 30), "CEST"),
        ("Europe/Berlin", datetime(2024, 10, 27, 2, 59, 59), "CEST"),
        ("Europe/Berlin", datetime(2024, 10, 27, 3, 0), "CET"),
        ("Europe/Berlin", datetime(2024, 10, 27, 3, 0, 1), "CET"),
        ("Europe/Berlin", datetime(2024, 10, 27, 4, 0), "CET"),
        # transition times from https://www.zeitverschiebung.net/de/timezone/america--new_york
        ("America/New_York", datetime(1970, 1, 1), "EST"),
        # Daylight Saving Time
        ("America/New_York", datetime(2024, 11, 3, 0, 0), "EDT"),
        ("America/New_York", datetime(2024, 11, 3, 1, 0), "EDT"),
        ("America/New_York", datetime(2024, 11, 3, 1, 59, 59), "EDT"),
        # 03.11.2024 2:00am -> 1:00am Standard
        # ("America/New_York", datetime(2024, 11, 3, 2, 0), "EDT"),
        ("America/New_York", datetime(2024, 11, 3, 2, 0, 1), "EST"),
        ("America/New_York", datetime(2024, 11, 3, 3, 0), "EST"),
        ("America/New_York", datetime(2025, 3, 9, 1, 0), "EST"),
        ("America/New_York", datetime(2025, 3, 9, 1, 59, 59), "EST"),
        ("America/New_York", datetime(2025, 3, 9, 2, 0), "EST"),
        # 09.03.2025 2:00am -> 3:00am Daylight Saving Time
        ("America/New_York", datetime(2025, 3, 9, 3, 0), "EDT"),
        ("America/New_York", datetime(2025, 3, 9, 3, 1, 1), "EDT"),
        ("America/New_York", datetime(2025, 3, 9, 4, 0), "EDT"),
    ],
)
def test_check_datetimes_around_transition_times(tzp, tzid, dt, tzname):
    """We should make sure than the datetimes with the generated timezones
    work as expected: They have the right UTC offset, dst and tzname.
    """
    message = f"{tzid}: {dt} (expected in {tzname})"
    expected_dt = tzp.localize(dt, tzid)
    component = Timezone.from_tzinfo(tzp.timezone(tzid))
    generated_tzinfo = component.to_tz(tzp, lookup_tzid=False)
    generated_dt = dt.replace(tzinfo=generated_tzinfo)
    print(generated_tzinfo)
    if tzp.uses_pytz():
        # generated_dt = generated_tzinfo.localize(dt)
        generated_dt = generated_tzinfo.normalize(generated_dt)
        if dt in (
            datetime(2024, 10, 27, 1, 0),
            datetime(2024, 11, 3, 1, 59, 59),
            datetime(2024, 11, 3, 1, 0),
            datetime(2024, 11, 3, 0, 0),
            datetime(2024, 10, 27, 2, 59, 59),
            datetime(2024, 10, 27, 2, 30),
            datetime(2024, 10, 27, 2, 0),
            datetime(2024, 10, 27, 1, 0),
        ):
            pytest.skip("We do not know how to do this. PYTZ-TODO")
    assert generated_dt.tzname() == expected_dt.tzname() == tzname, message
    assert generated_dt.dst() == expected_dt.dst(), message
    assert generated_dt.utcoffset() == expected_dt.utcoffset(), message


@pytest.mark.parametrize("uid", [0, 1, 2, 3])
def test_dateutil_timezone_when_time_is_going_backwards(calendars, tzp, uid):
    """When going from Daylight Saving Time to Standard Time, times can be ambiguous.
    For example, at 3:00 AM, the time falls back to 2:00 AM, repeating a full hour of times from 2:00 AM to 3:00 AM on the same date.

    By the RFC 5545, we cannot accommodate this case. All datetimes should
    be BEFORE the transition if ambiguous. However, dateutil can
    create a timezone that allows the event to be after this ambiguous time span, of course.

    Each event has its timezone saved in it.
    """
    cal: Calendar = calendars.issue_722_timezone_transition_ambiguity
    event: Event = cal.events[uid]
    expected_tzname = str(event["X-TZNAME"])
    actual_tzname = event.start.tzname()
    assert actual_tzname == expected_tzname, event["SUMMARY"]


def query_tzid(query: str, cal: Calendar) -> str:
    """The tzid from the query."""
    try:
        tzinfo = eval(query, {"cal": cal})  # noqa: S307
    except Exception as e:
        raise ValueError(query) from e
    return tzid_from_tzinfo(tzinfo)


# these are queries for all the places that have a TZID
# according to RFC 5545
queries = [
    "cal.events[0].start.tzinfo",  # DTSTART
    "cal.events[0].end.tzinfo",  # DTEND
    # EXDATE
    "cal.todos[0].end.tzinfo",  # DUE
    "cal.events[0].get('RDATE').dts[0].dt[0].tzinfo",  # RDATE
    "cal.events[1].get('RECURRENCE-ID').dt.tzinfo",  # RECURRENCE-ID
    "cal.events[2].get('RDATE')[0].dts[0].dt.tzinfo",  # RDATE multiple
    "cal.events[2].get('RDATE')[1].dts[0].dt.tzinfo",  # RDATE multiple
]


@pytest.mark.parametrize("query", queries)
def test_add_missing_timezones_to_example(calendars, query):
    """Add the missing timezones to the calendar."""
    cal = calendars.issue_722_missing_timezones
    tzid = query_tzid(query, cal)
    tzs = cal.get_missing_tzids()
    assert tzid in tzs


def test_queries_yield_unique_tzids(calendars):
    """We make sure each query tests a unique place to find for the algorithm."""
    cal = calendars.issue_722_missing_timezones
    tzids = set()
    for query in queries:
        tzid = query_tzid(query, cal)
        print(query, "->", tzid)
        tzids.add(tzid)
    assert len(tzids) == len(queries)


def test_we_do_not_miss_to_add_a_query(calendars):
    """Make sure all tzids are actually queried."""
    cal = calendars.issue_722_missing_timezones
    raw = cal.raw_ics.decode()
    ids = set(findall("TZID=([a-zA-Z_/+-]+)", raw))
    assert cal.get_used_tzids() == ids, "We find all tzids and they are unique."
    assert len(ids) == len(queries), "We query all the tzids."


def test_unknown_tzid(calendars):
    """If we have an unknown tzid with no timezone component."""
    cal = calendars.issue_722_missing_VTIMEZONE_custom
    assert "CUSTOM_tzid" in cal.get_used_tzids()
    assert "CUSTOM_tzid" in cal.get_missing_tzids()


def test_custom_timezone_is_found_and_used(calendars):
    """Check the custom timezone component is not missing."""
    cal = calendars.america_new_york
    assert "custom_America/New_York" in cal.get_used_tzids()
    assert "custom_America/New_York" not in cal.get_missing_tzids()


def test_not_missing_anything():
    """Check that no timezone is missing."""
    cal = Calendar()
    assert cal.get_missing_tzids() == set()


def test_utc_is_not_missing(calendars):
    """UTC should not be found missing."""
    cal = calendars.issue_722_missing_timezones
    assert "UTC" not in cal.get_missing_tzids()
    assert "UTC" not in cal.get_used_tzids()


def test_dateutil_timezone_is_not_found_with_tzname(calendars, no_pytz):
    """dateutil is an example of a timezone that has no tzid.

    In this test we make sure that the timezone is said to be missing.
    """
    cal: Calendar = calendars.america_new_york
    cal.subcomponents.remove(cal.timezones[0])
    assert cal.get_missing_tzids() == {"custom_America/New_York"}
    assert "dateutil" in repr(cal.events[0].start.tzinfo.__class__)


@pytest.mark.parametrize("tzname", ["America/New_York", "Arctic/Longyearbyen"])
# @pytest.mark.parametrize("component", ["STANDARD", "DAYLIGHT"])
def test_dateutil_timezone_is_matched_with_tzname(tzname):
    """dateutil is an example of a timezone that has no tzid.

    In this test we make sure that the timezone is matched by its
    tzname() in the timezone in the STANDARD and DAYLIGHT components.
    """
    cal = Calendar()
    event = Event()
    event.start = datetime(2024, 11, 12, tzinfo=gettz(tzname))
    print(dir(event.start.tzinfo))
    cal.add_component(event)
    assert cal.get_missing_tzids() == {tzname}
    cal.add_missing_timezones()
    assert cal.get_missing_tzids() == set()


def test_dateutil_timezone_is_also_added(calendars):
    """We find and add a dateutil timezone.

    This is important as we use those in the zoneinfo implementation.
    """


@pytest.mark.parametrize(
    "calendar",
    [
        "example",
        "america_new_york",  # custom
        "timezone_same_start",  # known tzid
        "period_with_timezone",  # known tzid
    ],
)
def test_timezone_is_not_missing(calendars, calendar):
    """Check that these calendars have no timezone missing."""
    cal: Calendar = calendars[calendar]
    timezones = cal.timezones[:]
    assert set() == cal.get_missing_tzids()
    cal.add_missing_timezones()
    assert set() == cal.get_missing_tzids()
    assert cal.timezones == timezones


def test_add_missing_known_timezones(calendars):
    """Add all timezones specified."""
    cal: Calendar = calendars.issue_722_missing_timezones
    assert len(cal.timezones) == 0
    cal.add_missing_timezones()
    assert len(cal.timezones) == len(queries), "all timezones are known"
    assert len(cal.get_missing_tzids()) == 0


def test_cannot_add_unknown_timezone(calendars):
    """I cannot add a timezone that is unknown."""
    cal: Calendar = calendars.issue_722_missing_VTIMEZONE_custom
    assert len(cal.timezones) == 0
    assert cal.get_missing_tzids() == {"CUSTOM_tzid"}
    cal.add_missing_timezones()
    assert cal.timezones == [], "we cannot add this timezone"
    assert cal.get_missing_tzids() == {"CUSTOM_tzid"}


def test_cannot_create_a_timezone_from_an_invalid_tzid():
    with pytest.raises(ValueError):
        Timezone.from_tzid("invalid/tzid")


def test_dates_before_and_after_are_considered():
    """When we add the timezones, we should check the calendar to see
    if all dates really occur in the span we use.

    We should also consider a huge default range.
    """
    pytest.skip("todo")


@pytest.mark.parametrize("tzid", zoneinfo.available_timezones() - {"Factory", "localtime"})
def test_we_can_identify_dateutil_timezones(tzid):
    """dateutil and others were badly supported.

    But if we know their shortcodes, we should be able to identify them.
    """
    tz = gettz(tzid)
    assert tz is None or tzid in tzids_from_tzinfo(tz)



"""Test VALUE parameter handling for Issue #349.

Tests that the VALUE parameter is properly set when adding properties with
Python native types (date, datetime, time, etc.) and properly used when
parsing iCalendar data.

Related Issues:
- #349: Adding an EXDATE with type date does not set the element's VALUE as DATE
- #187: Property values ignoring property parameters

Related PRs:
- #331: Fix VALUE parameter handling (2021, by thet)
- #196: Original attempt (2016, by stlaz)
"""

import datetime

from icalendar import Calendar, Event


class TestValueParameterEncoding:
    """Test that VALUE parameter is automatically set during encoding."""

    def test_exdate_with_date_sets_value_date(self):
        """Test that adding EXDATE with datetime.date sets VALUE=DATE."""
        ev = Event()
        ev.add("UID", "test-value-param-date")
        ev.add("DTSTAMP", datetime.datetime.now(datetime.timezone.utc))
        ev.add("SUMMARY", "EXDATE test")
        ev.add("DTSTART", datetime.date(2022, 1, 1))
        ev.add("RRULE", {"FREQ": "WEEKLY"})
        ev.add("EXDATE", datetime.date(2022, 1, 8))

        ical_str = ev.to_ical().decode()
        assert "EXDATE;VALUE=DATE:20220108" in ical_str

    def test_rdate_with_date_sets_value_date(self):
        """Test that adding RDATE with datetime.date sets VALUE=DATE."""
        ev = Event()
        ev.add("UID", "test-rdate-date")
        ev.add("DTSTAMP", datetime.datetime.now(datetime.timezone.utc))
        ev.add("DTSTART", datetime.date(2022, 1, 1))
        ev.add("RDATE", datetime.date(2022, 1, 15))

        ical_str = ev.to_ical().decode()
        assert "RDATE;VALUE=DATE:20220115" in ical_str

    def test_dtstart_with_date_sets_value_date(self):
        """Test that DTSTART with datetime.date sets VALUE=DATE."""
        ev = Event()
        ev.add("UID", "test-dtstart-date")
        ev.add("DTSTAMP", datetime.datetime.now(datetime.timezone.utc))
        ev.add("DTSTART", datetime.date(2022, 1, 1))

        ical_str = ev.to_ical().decode()
        assert "DTSTART;VALUE=DATE:20220101" in ical_str

    def test_dtstart_with_datetime_no_value_param(self):
        """Test that DTSTART with datetime.datetime doesn't set VALUE (default DATE-TIME)."""
        ev = Event()
        ev.add("UID", "test-dtstart-datetime")
        ev.add("DTSTAMP", datetime.datetime.now(datetime.timezone.utc))
        ev.add("DTSTART", datetime.datetime(2022, 1, 1, 10, 0, 0, tzinfo=datetime.timezone.utc))

        ical_str = ev.to_ical().decode()
        # DATE-TIME is the default, so VALUE parameter should not be explicitly set
        # (though some implementations may still add it)
        assert "DTSTART" in ical_str
        assert "20220101T100000Z" in ical_str

    def test_exdate_list_with_dates_sets_value_date(self):
        """Test that EXDATE list with datetime.date objects sets VALUE=DATE."""
        ev = Event()
        ev.add("UID", "test-exdate-list")
        ev.add("DTSTAMP", datetime.datetime.now(datetime.timezone.utc))
        ev.add("DTSTART", datetime.date(2022, 1, 1))
        ev.add("RRULE", {"FREQ": "DAILY"})
        # Add dates individually - lists are handled by vDDDLists
        ev.add("EXDATE", datetime.date(2022, 1, 5))
        ev.add("EXDATE", datetime.date(2022, 1, 10))

        ical_str = ev.to_ical().decode()
        assert "EXDATE;VALUE=DATE:20220105" in ical_str
        assert "EXDATE;VALUE=DATE:20220110" in ical_str


class TestValueParameterDecoding:
    """Test that VALUE parameter is properly used during decoding."""

    def test_parse_exdate_with_value_date(self):
        """Test parsing EXDATE with VALUE=DATE parameter."""
        ical_str = b"""BEGIN:VCALENDAR
VERSION:2.0
PRODID:-//Test//Test//EN
BEGIN:VEVENT
UID:test-parse-exdate-date
DTSTART;VALUE=DATE:20220101
EXDATE;VALUE=DATE:20220108
DTSTAMP:20220101T120000Z
SUMMARY:Test Event
END:VEVENT
END:VCALENDAR"""

        cal = Calendar.from_ical(ical_str)
        event = cal.walk("VEVENT")[0]
        exdate = event["EXDATE"]

        # EXDATE is always a vDDDLists, get the first item
        assert len(exdate.dts) == 1
        first_date = exdate.dts[0].dt
        # Should be parsed as a date, not datetime
        assert isinstance(first_date, datetime.date)
        assert not isinstance(first_date, datetime.datetime)
        assert first_date == datetime.date(2022, 1, 8)

    def test_parse_dtstart_with_value_date(self):
        """Test parsing DTSTART with VALUE=DATE parameter."""
        ical_str = b"""BEGIN:VCALENDAR
VERSION:2.0
PRODID:-//Test//Test//EN
BEGIN:VEVENT
UID:test-parse-dtstart-date
DTSTART;VALUE=DATE:20220101
DTSTAMP:20220101T120000Z
SUMMARY:Test Event
END:VEVENT
END:VCALENDAR"""

        cal = Calendar.from_ical(ical_str)
        event = cal.walk("VEVENT")[0]
        dtstart = event["DTSTART"]

        # Should be parsed as a date, not datetime
        assert isinstance(dtstart.dt, datetime.date)
        assert not isinstance(dtstart.dt, datetime.datetime)
        assert dtstart.dt == datetime.date(2022, 1, 1)

    def test_parse_dtstart_without_value_defaults_to_datetime(self):
        """Test that DTSTART without VALUE parameter defaults to DATE-TIME."""
        ical_str = b"""BEGIN:VCALENDAR
VERSION:2.0
PRODID:-//Test//Test//EN
BEGIN:VEVENT
UID:test-parse-dtstart-default
DTSTART:20220101T100000Z
DTSTAMP:20220101T120000Z
SUMMARY:Test Event
END:VEVENT
END:VCALENDAR"""

        cal = Calendar.from_ical(ical_str)
        event = cal.walk("VEVENT")[0]
        dtstart = event["DTSTART"]

        # Should be parsed as datetime
        assert isinstance(dtstart.dt, datetime.datetime)
        assert dtstart.dt == datetime.datetime(2022, 1, 1, 10, 0, 0, tzinfo=datetime.timezone.utc)


class TestValueParameterRoundTrip:
    """Test that VALUE parameter survives encoding/decoding round trips."""

    def test_date_roundtrip(self):
        """Test that date values maintain their type through encode/decode."""
        # Create event with date
        ev1 = Event()
        ev1.add("UID", "test-roundtrip")
        ev1.add("DTSTAMP", datetime.datetime.now(datetime.timezone.utc))
        ev1.add("DTSTART", datetime.date(2022, 1, 1))
        ev1.add("EXDATE", datetime.date(2022, 1, 8))

        # Encode to iCalendar format
        ical_bytes = ev1.to_ical()

        # Decode back
        cal = Calendar.from_ical(b"BEGIN:VCALENDAR\nVERSION:2.0\nPRODID:test\n" + ical_bytes + b"\nEND:VCALENDAR")
        ev2 = cal.walk("VEVENT")[0]

        # Check that types are preserved
        assert isinstance(ev2["DTSTART"].dt, datetime.date)
        assert not isinstance(ev2["DTSTART"].dt, datetime.datetime)
        assert ev2["DTSTART"].dt == datetime.date(2022, 1, 1)

        # EXDATE is always a vDDDLists, get the first item
        assert len(ev2["EXDATE"].dts) == 1
        exdate = ev2["EXDATE"].dts[0].dt
        assert isinstance(exdate, datetime.date)
        assert not isinstance(exdate, datetime.datetime)
        assert exdate == datetime.date(2022, 1, 8)


class TestExplicitValueParameter:
    """Test handling of explicitly provided VALUE parameters."""

    def test_explicit_value_param_overrides_inference(self):
        """Test that explicitly provided VALUE parameter is used."""
        ev = Event()
        ev.add("UID", "test-explicit-value")
        ev.add("DTSTAMP", datetime.datetime.now(datetime.timezone.utc))
        # Explicitly set VALUE=DATE
        ev.add("DTSTART", datetime.date(2022, 1, 1), parameters={"VALUE": "DATE"})

        ical_str = ev.to_ical().decode()
        assert "VALUE=DATE" in ical_str

    def test_parameters_preserved(self):
        """Test that other parameters are preserved when VALUE is added."""
        ev = Event()
        ev.add("UID", "test-params-preserved")
        ev.add("DTSTAMP", datetime.datetime.now(datetime.timezone.utc))
        ev.add("DTSTART", datetime.datetime(2022, 1, 1, 10, 0, 0), parameters={"TZID": "America/New_York"})

        dtstart = ev["DTSTART"]
        assert "TZID" in dtstart.params
        assert dtstart.params["TZID"] == "America/New_York"



"""Conferences according to Section 5.11 of :rfc:`7986`."""

from __future__ import annotations

from dataclasses import dataclass

from icalendar.prop import vUri


@dataclass
class Conference:
    """Conferences according to Section 5.11 of :rfc:`7986`.

    Purpose:
        Information for accessing a conferencing system.

    Conformance:
        This property can be specified multiple times in a
        "VEVENT" or "VTODO" calendar component.

    Description:
        This property specifies information for accessing a
        conferencing system for attendees of a meeting or task.  This
        might be for a telephone-based conference number dial-in with
        access codes included (such as a tel: URI :rfc:`3966` or a sip: or
        sips: URI :rfc:`3261`), for a web-based video chat (such as an http:
        or https: URI :rfc:`7230`), or for an instant messaging group chat
        room (such as an xmpp: URI :rfc:`5122`).  If a specific URI for a
        conferencing system is not available, a data: URI :rfc:`2397`
        containing a text description can be used.

        A conference system can be a bidirectional communication channel
        or a uni-directional "broadcast feed".

        The "FEATURE" property parameter is used to describe the key
        capabilities of the conference system to allow a client to choose
        the ones that give the required level of interaction from a set of
        multiple properties.

        The "LABEL" property parameter is used to convey additional
        details on the use of the URI.  For example, the URIs or access
        codes for the moderator and attendee of a teleconference system
        could be different, and the "LABEL" property parameter could be
        used to "tag" each "CONFERENCE" property to indicate which is
        which.

        The "LANGUAGE" property parameter can be used to specify the
        language used for text values used with this property (as per
        Section 3.2.10 of :rfc:`5545`).

    Example:
        The following are examples of this property:

        .. code-block:: text

            CONFERENCE;VALUE=URI;FEATURE=PHONE,MODERATOR;
             LABEL=Moderator dial-in:tel:+1-412-555-0123,,,654321
            CONFERENCE;VALUE=URI;FEATURE=PHONE;
             LABEL=Attendee dial-in:tel:+1-412-555-0123,,,555123
            CONFERENCE;VALUE=URI;FEATURE=PHONE;
             LABEL=Attendee dial-in:tel:+1-888-555-0456,,,555123
            CONFERENCE;VALUE=URI;FEATURE=CHAT;
             LABEL=Chat room:xmpp:chat-123@conference.example.com
            CONFERENCE;VALUE=URI;FEATURE=AUDIO,VIDEO;
             LABEL=Attendee dial-in:https://chat.example.com/audio?id=123456
    """

    # see https://stackoverflow.com/a/18348004/1320237
    uri: str
    feature: list[str] | str | None = None
    label: str | None = None
    language: str | None = None

    @classmethod
    def from_uri(cls, uri: vUri | str):
        """Create a Conference from a URI."""
        params = uri.params if isinstance(uri, vUri) else {}
        return cls(
            uri,
            feature=params.get("feature"),
            label=params.get("label"),
            language=params.get("language"),
        )

    def to_uri(self) -> vUri:
        """Convert the Conference to a vUri."""

        def normalize(value: str | list[str] | None) -> str | None:
            if value is None or (isinstance(value, list) and len(value) == 0):
                return None
            if isinstance(value, str):
                return value
            return ",".join(value)

        params: dict[str, str | None] = {
            "FEATURE": normalize(self.feature),
            "LABEL": normalize(self.label),
            "LANGUAGE": normalize(self.language),
            "VALUE": "URI",
        }

        filtered_params: dict[str, str] = {
            key: value for key, value in params.items() if value is not None
        }
        return vUri(self.uri, params=filtered_params)


__all__ = ["Conference"]



"""This file tests the source code provided by the documentation.

See
- doctest documentation: https://docs.python.org/3/library/doctest.html
- Issue 443: https://github.com/collective/icalendar/issues/443

This file should be tests, too:

    >>> print("Hello World!")
    Hello World!

"""

import doctest
import importlib
import sys
from pathlib import Path

import pytest

HERE = Path(__file__ or ".").absolute().parent
ICALENDAR_PATH = HERE / ".."

PYTHON_FILES = [
    file
    for file in ICALENDAR_PATH.glob("**/*.py")
    if "fuzzing" not in file.relative_to(ICALENDAR_PATH).parts
]

MODULE_NAMES = [
    "icalendar"
    + str(python_file)[len(str(ICALENDAR_PATH)) : -3]
    .replace("\\", "/")
    .replace("/", ".")
    for python_file in PYTHON_FILES
]


def test_this_module_is_among_them():
    assert __name__ in MODULE_NAMES


@pytest.mark.parametrize("module_name", MODULE_NAMES)
def test_docstring_of_python_file(module_name, env_for_doctest):
    """This test runs doctest on the Python module."""
    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError as e:
        if e.name == "pytz":
            pytest.skip("pytz is not installed, skipping this module.")
        raise
    test_result = doctest.testmod(module, name=module_name, globs=env_for_doctest)
    assert test_result.failed == 0, f"{test_result.failed} errors in {module_name}"


# This collection needs to exclude .tox and other subdirectories
REPOSITORY = HERE.parent.parent.parent
DOCUMENTATION_ROOT = REPOSITORY / "docs"
IGNORED_DOCUMENTATION_FILES = [
    DOCUMENTATION_ROOT / "contribute" / "documentation.rst"  # docstring
]
DOCUMENT_PATHS = [
    file
    for file in REPOSITORY.glob("**/*.rst")
    if file not in IGNORED_DOCUMENTATION_FILES
]


@pytest.mark.parametrize(
    "filename",
    [
        "README.rst",
        "index.rst",
    ],
)
def test_files_is_included(filename):
    assert any(path.name == filename for path in DOCUMENT_PATHS)


@pytest.mark.parametrize("document", DOCUMENT_PATHS)
def test_documentation_file(document, zoneinfo_only, env_for_doctest, tzp):
    """This test runs doctest on a documentation file.

    functions are also replaced to work.
    """
    try:
        import pytz  # noqa: F401
    except ImportError:
        pytest.skip("pytz not installed, skipping this file.")
    try:
        # set raise_on_error to False if you wand to see the error for debug
        test_result = doctest.testfile(
            document, module_relative=False, globs=env_for_doctest, raise_on_error=False
        )
    finally:
        tzp.use_zoneinfo()
    assert test_result.failed == 0, (
        f"{test_result.failed} errors in {Path(document).name}"
    )


def test_can_import_zoneinfo(env_for_doctest):
    """Allow importing zoneinfo for tests."""
    assert "zoneinfo" in sys.modules



""":rfc:`5545` components for timezone information."""

from __future__ import annotations

from collections import defaultdict
from datetime import date, datetime, timedelta, tzinfo
from typing import TYPE_CHECKING, Optional

import dateutil.rrule
import dateutil.tz

from icalendar.attr import (
    create_single_property,
    exdates_property,
    rdates_property,
    rrules_property,
)
from icalendar.cal.component import Component
from icalendar.cal.examples import get_example
from icalendar.prop import tzid_from_tzinfo, vUTCOffset
from icalendar.timezone import TZP, tzp
from icalendar.tools import is_date, to_datetime

if TYPE_CHECKING:
    from icalendar.cal.calendar import Calendar


class Timezone(Component):
    """
    A "VTIMEZONE" calendar component is a grouping of component
    properties that defines a time zone. It is used to describe the
    way in which a time zone changes its offset from UTC over time.
    """

    subcomponents: list[TimezoneStandard | TimezoneDaylight]

    name = "VTIMEZONE"
    canonical_order = ("TZID",)
    required = ("TZID",)  # it also requires one of components DAYLIGHT and STANDARD
    singletons = (
        "TZID",
        "LAST-MODIFIED",
        "TZURL",
    )

    DEFAULT_FIRST_DATE = date(1970, 1, 1)
    DEFAULT_LAST_DATE = date(2038, 1, 1)

    @classmethod
    def example(cls, name: str = "pacific_fiji") -> Calendar:
        """Return the timezone example with the given name."""
        return cls.from_ical(get_example("timezones", name))

    @staticmethod
    def _extract_offsets(component: TimezoneDaylight | TimezoneStandard, tzname: str):
        """extract offsets and transition times from a VTIMEZONE component
        :param component: a STANDARD or DAYLIGHT component
        :param tzname: the name of the zone
        """
        offsetfrom = component.TZOFFSETFROM
        offsetto = component.TZOFFSETTO
        dtstart = component.DTSTART

        # offsets need to be rounded to the next minute, we might loose up
        # to 30 seconds accuracy, but it can't be helped (datetime
        # supposedly cannot handle smaller offsets)
        offsetto_s = int((offsetto.seconds + 30) / 60) * 60
        offsetto = timedelta(days=offsetto.days, seconds=offsetto_s)
        offsetfrom_s = int((offsetfrom.seconds + 30) / 60) * 60
        offsetfrom = timedelta(days=offsetfrom.days, seconds=offsetfrom_s)

        # expand recurrences
        if "RRULE" in component:
            # to be paranoid about correct weekdays
            # evaluate the rrule with the current offset
            tzi = dateutil.tz.tzoffset("(offsetfrom)", offsetfrom)
            rrstart = dtstart.replace(tzinfo=tzi)

            rrulestr = component["RRULE"].to_ical().decode("utf-8")
            rrule = dateutil.rrule.rrulestr(rrulestr, dtstart=rrstart)
            tzp.fix_rrule_until(rrule, component["RRULE"])

            # constructing the timezone requires UTC transition times.
            # here we construct local times without tzinfo, the offset to UTC
            # gets subtracted in to_tz().
            transtimes = [dt.replace(tzinfo=None) for dt in rrule]

        # or rdates
        elif "RDATE" in component:
            if not isinstance(component["RDATE"], list):
                rdates = [component["RDATE"]]
            else:
                rdates = component["RDATE"]
            transtimes = [dtstart] + [leaf.dt for tree in rdates for leaf in tree.dts]
        else:
            transtimes = [dtstart]

        transitions = [
            (transtime, offsetfrom, offsetto, tzname) for transtime in set(transtimes)
        ]

        if component.name == "STANDARD":
            is_dst = 0
        elif component.name == "DAYLIGHT":
            is_dst = 1
        return is_dst, transitions

    @staticmethod
    def _make_unique_tzname(tzname, tznames):
        """
        :param tzname: Candidate tzname
        :param tznames: Other tznames
        """
        # TODO better way of making sure tznames are unique
        while tzname in tznames:
            tzname += "_1"
        tznames.add(tzname)
        return tzname

    def to_tz(self, tzp: TZP = tzp, lookup_tzid: bool = True):  # noqa: FBT001
        """convert this VTIMEZONE component to a timezone object

        :param tzp: timezone provider to use
        :param lookup_tzid: whether to use the TZID property to look up existing
                            timezone definitions with tzp.
                            If it is False, a new timezone will be created.
                            If it is True, the existing timezone will be used
                            if it exists, otherwise a new timezone will be created.
        """
        if lookup_tzid:
            tz = tzp.timezone(self.tz_name)
            if tz is not None:
                return tz
        return tzp.create_timezone(self)

    @property
    def tz_name(self) -> str:
        """Return the name of the timezone component.

        Please note that the names of the timezone are different from this name
        and may change with winter/summer time.
        """
        try:
            return str(self["TZID"])
        except UnicodeEncodeError:
            return self["TZID"].encode("ascii", "replace")

    def get_transitions(
        self,
    ) -> tuple[list[datetime], list[tuple[timedelta, timedelta, str]]]:
        """Return a tuple of (transition_times, transition_info)

        - transition_times = [datetime, ...]
        - transition_info = [(TZOFFSETTO, dts_offset, tzname)]

        """
        zone = self.tz_name
        transitions = []
        dst = {}
        tznames = set()
        for component in self.walk():
            if isinstance(component, Timezone):
                continue
            if is_date(component["DTSTART"].dt):
                component.DTSTART = to_datetime(component["DTSTART"].dt)
            assert isinstance(component["DTSTART"].dt, datetime), (
                "VTIMEZONEs sub-components' DTSTART must be of type datetime, not date"
            )
            try:
                tzname = str(component["TZNAME"])
            except UnicodeEncodeError:
                tzname = component["TZNAME"].encode("ascii", "replace")
                tzname = self._make_unique_tzname(tzname, tznames)
            except KeyError:
                # for whatever reason this is str/unicode
                tzname = (
                    f"{zone}_{component['DTSTART'].to_ical().decode('utf-8')}_"
                    f"{component['TZOFFSETFROM'].to_ical()}_"
                    f"{component['TZOFFSETTO'].to_ical()}"
                )
                tzname = self._make_unique_tzname(tzname, tznames)

            dst[tzname], component_transitions = self._extract_offsets(
                component, tzname
            )
            transitions.extend(component_transitions)

        transitions.sort()
        transition_times = [
            transtime - osfrom for transtime, osfrom, _, _ in transitions
        ]

        # transition_info is a list with tuples in the format
        # (utcoffset, dstoffset, name)
        # dstoffset = 0, if current transition is to standard time
        #           = this_utcoffset - prev_standard_utcoffset, otherwise
        transition_info = []
        for num, (_transtime, osfrom, osto, name) in enumerate(transitions):
            dst_offset = False
            if not dst[name]:
                dst_offset = timedelta(seconds=0)
            else:
                # go back in time until we find a transition to dst
                for index in range(num - 1, -1, -1):
                    if not dst[transitions[index][3]]:  # [3] is the name
                        dst_offset = osto - transitions[index][2]  # [2] is osto
                        break
                # when the first transition is to dst, we didn't find anything
                # in the past, so we have to look into the future
                if not dst_offset:
                    for index in range(num, len(transitions)):
                        if not dst[transitions[index][3]]:  # [3] is the name
                            dst_offset = osto - transitions[index][2]  # [2] is osto
                            break
                # If we still haven't found a STANDARD transition
                # (only DAYLIGHT exists), calculate dst_offset as the
                # difference from TZOFFSETFROM. Handles Issue #321.
                if dst_offset is False:
                    dst_offset = osto - osfrom
            transition_info.append((osto, dst_offset, name))
        return transition_times, transition_info

    # binary search
    _from_tzinfo_skip_search = [
        timedelta(days=days) for days in (64, 32, 16, 8, 4, 2, 1)
    ] + [
        # we know it happens in the night usually around 1am
        timedelta(hours=4),
        timedelta(hours=1),
        # adding some minutes and seconds for faster search
        timedelta(minutes=20),
        timedelta(minutes=5),
        timedelta(minutes=1),
        timedelta(seconds=20),
        timedelta(seconds=5),
        timedelta(seconds=1),
    ]

    @classmethod
    def from_tzinfo(
        cls,
        timezone: tzinfo,
        tzid: Optional[str] = None,
        first_date: date = DEFAULT_FIRST_DATE,
        last_date: date = DEFAULT_LAST_DATE,
    ) -> Timezone:
        """Return a VTIMEZONE component from a timezone object.

        This works with pytz and zoneinfo and any other timezone.
        The offsets are calculated from the tzinfo object.

        Parameters:

        :param tzinfo: the timezone object
        :param tzid: the tzid for this timezone. If None, it will be extracted from the tzinfo.
        :param first_date: a datetime that is earlier than anything that happens in the calendar
        :param last_date: a datetime that is later than anything that happens in the calendar
        :raises ValueError: If we have no tzid and cannot extract one.

        .. note::
            This can take some time. Please cache the results.
        """  # noqa: E501
        if tzid is None:
            tzid = tzid_from_tzinfo(timezone)
            if tzid is None:
                raise ValueError(
                    f"Cannot get TZID from {timezone}. Please set the tzid parameter."
                )
        normalize = getattr(timezone, "normalize", lambda dt: dt)  # pytz compatibility
        first_datetime = datetime(first_date.year, first_date.month, first_date.day)  # noqa: DTZ001
        last_datetime = datetime(last_date.year, last_date.month, last_date.day)  # noqa: DTZ001
        if hasattr(timezone, "localize"):  # pytz compatibility
            first_datetime = timezone.localize(first_datetime)
            last_datetime = timezone.localize(last_datetime)
        else:
            first_datetime = first_datetime.replace(tzinfo=timezone)
            last_datetime = last_datetime.replace(tzinfo=timezone)
        # from, to, tzname, is_standard -> start
        offsets: dict[
            tuple[Optional[timedelta], timedelta, str, bool], list[datetime]
        ] = defaultdict(list)
        start = first_datetime
        offset_to = None
        while start < last_datetime:
            offset_from = offset_to
            end = start
            offset_to = end.utcoffset()
            for add_offset in cls._from_tzinfo_skip_search:
                last_end = end  # we need to save this as we might be left and right of the time change  # noqa: E501
                end = normalize(end + add_offset)
                try:
                    while end.utcoffset() == offset_to:
                        last_end = end
                        end = normalize(end + add_offset)
                except OverflowError:
                    # zoninfo does not go all the way
                    break
                # retract if we overshoot
                end = last_end
            # Now, start (inclusive) -> end (exclusive) are one timezone
            is_standard = start.dst() == timedelta()
            name = start.tzname()
            if name is None:
                name = str(offset_to)
            key = (offset_from, offset_to, name, is_standard)
            # first_key = (None,) + key[1:]
            # if first_key in offsets:
            #     # remove the first one and claim it changes at that day
            #     offsets[first_key] = offsets.pop(first_key)
            offsets[key].append(start.replace(tzinfo=None))
            start = normalize(end + cls._from_tzinfo_skip_search[-1])
        tz = cls()
        tz.add("TZID", tzid)
        tz.add("COMMENT", f"This timezone only works from {first_date} to {last_date}.")
        for (offset_from, offset_to, tzname, is_standard), starts in offsets.items():
            first_start = min(starts)
            starts.remove(first_start)
            if first_start.date() == last_date:
                first_start = datetime(last_date.year, last_date.month, last_date.day)  # noqa: DTZ001
            subcomponent = TimezoneStandard() if is_standard else TimezoneDaylight()
            if offset_from is None:
                offset_from = offset_to  # noqa: PLW2901
            subcomponent.TZOFFSETFROM = offset_from
            subcomponent.TZOFFSETTO = offset_to
            subcomponent.add("TZNAME", tzname)
            subcomponent.DTSTART = first_start
            if starts:
                subcomponent.add("RDATE", starts)
            tz.add_component(subcomponent)
        return tz

    @classmethod
    def from_tzid(
        cls,
        tzid: str,
        tzp: TZP = tzp,
        first_date: date = DEFAULT_FIRST_DATE,
        last_date: date = DEFAULT_LAST_DATE,
    ) -> Timezone:
        """Create a VTIMEZONE from a tzid like ``"Europe/Berlin"``.

        :param tzid: the id of the timezone
        :param tzp: the timezone provider
        :param first_date: a datetime that is earlier than anything
            that happens in the calendar
        :param last_date: a datetime that is later than anything
            that happens in the calendar
        :raises ValueError: If the tzid is unknown.

        >>> from icalendar import Timezone
        >>> tz = Timezone.from_tzid("Europe/Berlin")
        >>> print(tz.to_ical()[:36])
        BEGIN:VTIMEZONE
        TZID:Europe/Berlin

        .. note::
            This can take some time. Please cache the results.
        """
        tz = tzp.timezone(tzid)
        if tz is None:
            raise ValueError(f"Unkown timezone {tzid}.")
        return cls.from_tzinfo(tz, tzid, first_date, last_date)

    @property
    def standard(self) -> list[TimezoneStandard]:
        """The STANDARD subcomponents as a list."""
        return self.walk("STANDARD")

    @property
    def daylight(self) -> list[TimezoneDaylight]:
        """The DAYLIGHT subcomponents as a list.

        These are for the daylight saving time.
        """
        return self.walk("DAYLIGHT")


class TimezoneStandard(Component):
    """
    The "STANDARD" sub-component of "VTIMEZONE" defines the standard
    time offset from UTC for a time zone. It represents a time zone's
    standard time, typically used during winter months in locations
    that observe Daylight Saving Time.
    """

    name = "STANDARD"
    required = ("DTSTART", "TZOFFSETTO", "TZOFFSETFROM")
    singletons = (
        "DTSTART",
        "TZOFFSETTO",
        "TZOFFSETFROM",
    )
    multiple = ("COMMENT", "RDATE", "TZNAME", "RRULE", "EXDATE")

    DTSTART = create_single_property(
        "DTSTART",
        "dt",
        (datetime,),
        datetime,
        """The mandatory "DTSTART" property gives the effective onset date
        and local time for the time zone sub-component definition.
        "DTSTART" in this usage MUST be specified as a date with a local
        time value.""",
    )
    TZOFFSETTO = create_single_property(
        "TZOFFSETTO",
        "td",
        (timedelta,),
        timedelta,
        """The mandatory "TZOFFSETTO" property gives the UTC offset for the
        time zone sub-component (Standard Time or Daylight Saving Time)
        when this observance is in use.
        """,
        vUTCOffset,
    )
    TZOFFSETFROM = create_single_property(
        "TZOFFSETFROM",
        "td",
        (timedelta,),
        timedelta,
        """The mandatory "TZOFFSETFROM" property gives the UTC offset that is
        in use when the onset of this time zone observance begins.
        "TZOFFSETFROM" is combined with "DTSTART" to define the effective
        onset for the time zone sub-component definition.  For example,
        the following represents the time at which the observance of
        Standard Time took effect in Fall 1967 for New York City:

            DTSTART:19671029T020000
            TZOFFSETFROM:-0400
        """,
        vUTCOffset,
    )
    rdates = rdates_property
    exdates = exdates_property
    rrules = rrules_property


class TimezoneDaylight(Component):
    """
    The "DAYLIGHT" sub-component of "VTIMEZONE" defines the daylight
    saving time offset from UTC for a time zone. It represents a time
    zone's daylight saving time, typically used during summer months
    in locations that observe Daylight Saving Time.
    """

    name = "DAYLIGHT"
    required = TimezoneStandard.required
    singletons = TimezoneStandard.singletons
    multiple = TimezoneStandard.multiple

    DTSTART = TimezoneStandard.DTSTART
    TZOFFSETTO = TimezoneStandard.TZOFFSETTO
    TZOFFSETFROM = TimezoneStandard.TZOFFSETFROM

    rdates = rdates_property
    exdates = exdates_property
    rrules = rrules_property


__all__ = ["Timezone", "TimezoneDaylight", "TimezoneStandard"]



"""Errors thrown by icalendar."""

from __future__ import annotations

import contextlib


class InvalidCalendar(ValueError):
    """The calendar given is not valid.

    This calendar does not conform with RFC 5545 or breaks other RFCs.
    """


class IncompleteComponent(ValueError):
    """The component is missing attributes.

    The attributes are not required, otherwise this would be
    an InvalidCalendar. But in order to perform calculations,
    this attribute is required.

    This error is not raised in the UPPERCASE properties like .DTSTART,
    only in the lowercase computations like .start.
    """


class IncompleteAlarmInformation(ValueError):
    """The alarms cannot be calculated yet because information is missing."""


class LocalTimezoneMissing(IncompleteAlarmInformation):
    """We are missing the local timezone to compute the value.

    Use Alarms.set_local_timezone().
    """


class ComponentEndMissing(IncompleteAlarmInformation):
    """We are missing the end of a component that the alarm is for.

    Use Alarms.set_end().
    """


class ComponentStartMissing(IncompleteAlarmInformation):
    """We are missing the start of a component that the alarm is for.

    Use Alarms.set_start().
    """


class FeatureWillBeRemovedInFutureVersion(DeprecationWarning):
    """This feature will be removed in a future version."""


def _repr_index(index: str | int) -> str:
    """Create a JSON compatible representation for the index."""
    if isinstance(index, str):
        return f'"{index}"'
    return str(index)


class JCalParsingError(ValueError):
    """Could not parse a part of the JCal."""

    _default_value = object()

    def __init__(
        self,
        message: str,
        parser: str | type = "",
        path: list[str | int] | None | str | int = None,
        value: object = _default_value,
    ):
        """Create a new JCalParsingError."""
        self.path = self._get_path(path)
        if not isinstance(parser, str):
            parser = parser.__name__
        self.parser = parser
        self.message = message
        self.value = value
        full_message = message
        repr_path = ""
        if self.path:
            repr_path = "".join([f"[{_repr_index(index)}]" for index in self.path])
            full_message = f"{repr_path}: {full_message}"
            repr_path += " "
        if parser:
            full_message = f"{repr_path}in {parser}: {message}"
        if value is not self._default_value:
            full_message += f" Got value: {value!r}"
        super().__init__(full_message)

    @classmethod
    @contextlib.contextmanager
    def reraise_with_path_added(cls, *path_components: int | str):
        """Automatically re-raise the exception with path components added.

        Raises:
            JCalParsingError: If there was an exception in the context.
        """
        try:
            yield
        except JCalParsingError as e:
            raise cls(
                path=list(path_components) + e.path,
                parser=e.parser,
                message=e.message,
                value=e.value,
            ).with_traceback(e.__traceback__) from e

    @staticmethod
    def _get_path(path: list[str | int] | None | str | int) -> list[str | int]:
        """Return the path as a list."""
        if path is None:
            path = []
        elif not isinstance(path, list):
            path = [path]
        return path

    @classmethod
    def validate_property(
        cls,
        jcal_property,
        parser: str | type,
        path: list[str | int] | None | str | int = None,
    ):
        """Validate a jCal property.

        Raises:
            JCalParsingError: if the property is not valid.
        """
        path = cls._get_path(path)
        if not isinstance(jcal_property, list) or len(jcal_property) < 4:
            raise JCalParsingError(
                "The property must be a list with at least 4 items.",
                parser,
                path,
                value=jcal_property,
            )
        if not isinstance(jcal_property[0], str):
            raise JCalParsingError(
                "The name must be a string.", parser, path + [0], value=jcal_property[0]
            )
        if not isinstance(jcal_property[1], dict):
            raise JCalParsingError(
                "The parameters must be a mapping.",
                parser,
                path + [1],
                value=jcal_property[1],
            )
        if not isinstance(jcal_property[2], str):
            raise JCalParsingError(
                "The VALUE parameter must be a string.",
                parser,
                path + [2],
                value=jcal_property[2],
            )

    _type_names = {
        str: "a string",
        int: "an integer",
        float: "a float",
        bool: "a boolean",
    }

    @classmethod
    def validate_value_type(
        cls,
        jcal,
        expected_type: type[str | int | float | bool]
        | tuple[type[str | int | float | bool], ...],
        parser: str | type = "",
        path: list[str | int] | None | str | int = None,
    ):
        """Validate the type of a jCal value."""
        if not isinstance(jcal, expected_type):
            type_name = (
                cls._type_names[expected_type]
                if isinstance(expected_type, type)
                else " or ".join(cls._type_names[t] for t in expected_type)
            )
            raise cls(
                f"The value must be {type_name}.",
                parser=parser,
                value=jcal,
                path=path,
            )

    @classmethod
    def validate_list_type(
        cls,
        jcal,
        expected_type: type[str | int | float | bool],
        parser: str | type = "",
        path: list[str | int] | None | str | int = None,
    ):
        """Validate the type of each item in a jCal list."""
        path = cls._get_path(path)
        if not isinstance(jcal, list):
            raise cls(
                "The value must be a list.",
                parser=parser,
                value=jcal,
                path=path,
            )
        for index, item in enumerate(jcal):
            if not isinstance(item, expected_type):
                type_name = cls._type_names[expected_type]
                raise cls(
                    f"Each item in the list must be {type_name}.",
                    parser=parser,
                    value=item,
                    path=path + [index],
                )


__all__ = [
    "ComponentEndMissing",
    "ComponentStartMissing",
    "FeatureWillBeRemovedInFutureVersion",
    "IncompleteAlarmInformation",
    "IncompleteComponent",
    "InvalidCalendar",
    "JCalParsingError",
    "LocalTimezoneMissing",
]



"""This module parses and generates contentlines as defined in RFC 5545
(iCalendar), but will probably work for other MIME types with similar syntax.
Eg. RFC 2426 (vCard)

It is stupid in the sense that it treats the content purely as strings. No type
conversion is attempted.
"""

from __future__ import annotations

import functools
import os
import re
from datetime import datetime, time
from typing import TYPE_CHECKING, Any, Callable

from icalendar.caselessdict import CaselessDict
from icalendar.error import JCalParsingError
from icalendar.parser_tools import (
    DEFAULT_ENCODING,
    ICAL_TYPE,
    SEQUENCE_TYPES,
    to_unicode,
)
from icalendar.timezone.tzid import tzid_from_dt

if TYPE_CHECKING:
    from icalendar.enums import VALUE
    from icalendar.prop import VPROPERTY


def escape_char(text):
    """Format value according to iCalendar TEXT escaping rules."""
    assert isinstance(text, (str, bytes))
    # NOTE: ORDER MATTERS!
    return (
        text.replace(r"\N", "\n")
        .replace("\\", "\\\\")
        .replace(";", r"\;")
        .replace(",", r"\,")
        .replace("\r\n", r"\n")
        .replace("\n", r"\n")
    )


def unescape_char(text):
    assert isinstance(text, (str, bytes))
    # NOTE: ORDER MATTERS!
    if isinstance(text, str):
        return (
            text.replace("\\N", "\\n")
            .replace("\r\n", "\n")
            .replace("\\n", "\n")
            .replace("\\,", ",")
            .replace("\\;", ";")
            .replace("\\\\", "\\")
        )
    if isinstance(text, bytes):
        return (
            text.replace(b"\\N", b"\\n")
            .replace(b"\r\n", b"\n")
            .replace(b"\\n", b"\n")
            .replace(b"\\,", b",")
            .replace(b"\\;", b";")
            .replace(b"\\\\", b"\\")
        )
    return None


def foldline(line, limit=75, fold_sep="\r\n "):
    """Make a string folded as defined in RFC5545
    Lines of text SHOULD NOT be longer than 75 octets, excluding the line
    break.  Long content lines SHOULD be split into a multiple line
    representations using a line "folding" technique.  That is, a long
    line can be split between any two characters by inserting a CRLF
    immediately followed by a single linear white-space character (i.e.,
    SPACE or HTAB).
    """
    assert isinstance(line, str)
    assert "\n" not in line

    # Use a fast and simple variant for the common case that line is all ASCII.
    try:
        line.encode("ascii")
    except (UnicodeEncodeError, UnicodeDecodeError):
        pass
    else:
        return fold_sep.join(
            line[i : i + limit - 1] for i in range(0, len(line), limit - 1)
        )

    ret_chars = []
    byte_count = 0
    for char in line:
        char_byte_len = len(char.encode(DEFAULT_ENCODING))
        byte_count += char_byte_len
        if byte_count >= limit:
            ret_chars.append(fold_sep)
            byte_count = char_byte_len
        ret_chars.append(char)

    return "".join(ret_chars)


#################################################################
# Property parameter stuff


def param_value(value, always_quote=False):
    """Returns a parameter value."""
    if isinstance(value, SEQUENCE_TYPES):
        return q_join(map(rfc_6868_escape, value), always_quote=always_quote)
    if isinstance(value, str):
        return dquote(rfc_6868_escape(value), always_quote=always_quote)
    return dquote(rfc_6868_escape(value.to_ical().decode(DEFAULT_ENCODING)))


# Could be improved

# [\w-] because of the iCalendar RFC
# . because of the vCard RFC
NAME = re.compile(r"[\w.-]+")

UNSAFE_CHAR = re.compile('[\x00-\x08\x0a-\x1f\x7f",:;]')
QUNSAFE_CHAR = re.compile('[\x00-\x08\x0a-\x1f\x7f"]')
FOLD = re.compile(b"(\r?\n)+[ \t]")
UFOLD = re.compile("(\r?\n)+[ \t]")
NEWLINE = re.compile(r"\r?\n")


def validate_token(name):
    match = NAME.findall(name)
    if len(match) == 1 and name == match[0]:
        return
    raise ValueError(name)


def validate_param_value(value, quoted=True):
    validator = QUNSAFE_CHAR if quoted else UNSAFE_CHAR
    if validator.findall(value):
        raise ValueError(value)


# chars presence of which in parameter value will be cause the value
# to be enclosed in double-quotes
QUOTABLE = re.compile("[,;:’]")  # noqa: RUF001


def dquote(val, always_quote=False):
    """Enclose parameter values containing [,;:] in double quotes."""
    # a double-quote character is forbidden to appear in a parameter value
    # so replace it with a single-quote character
    val = val.replace('"', "'")
    if QUOTABLE.search(val) or always_quote:
        return f'"{val}"'
    return val


# parsing helper
def q_split(st, sep=",", maxsplit=-1):
    """Splits a string on char, taking double (q)uotes into considderation."""
    if maxsplit == 0:
        return [st]

    result = []
    cursor = 0
    length = len(st)
    inquote = 0
    splits = 0
    for i, ch in enumerate(st):
        if ch == '"':
            inquote = not inquote
        if not inquote and ch == sep:
            result.append(st[cursor:i])
            cursor = i + 1
            splits += 1
        if i + 1 == length or splits == maxsplit:
            result.append(st[cursor:])
            break
    return result


def q_join(lst, sep=",", always_quote=False):
    """Joins a list on sep, quoting strings with QUOTABLE chars."""
    return sep.join(dquote(itm, always_quote=always_quote) for itm in lst)


def single_string_parameter(func: Callable | None = None, upper=False):
    """Create a parameter getter/setter for a single string parameter.

    Args:
        upper: Convert the value to uppercase
        func: The function to decorate.

    Returns:
        The property for the parameter or a decorator for the parameter
        if func is ``None``.
    """

    def decorator(func):
        name = func.__name__

        @functools.wraps(func)
        def fget(self: Parameters):
            """Get the value."""
            value = self.get(name)
            if value is not None and upper:
                value = value.upper()
            return value

        def fset(self: Parameters, value: str | None):
            """Set the value"""
            if value is None:
                fdel(self)
            else:
                if upper:
                    value = value.upper()
                self[name] = value

        def fdel(self: Parameters):
            """Delete the value."""
            self.pop(name, None)

        return property(fget, fset, fdel, doc=func.__doc__)

    if func is None:
        return decorator
    return decorator(func)


class Parameters(CaselessDict):
    """Parser and generator of Property parameter strings.

    It knows nothing of datatypes.
    Its main concern is textual structure.

    Examples:

        Modify parameters:

        .. code-block:: pycon

            >>> from icalendar import Parameters
            >>> params = Parameters()
            >>> params['VALUE'] = 'TEXT'
            >>> params.value
            'TEXT'
            >>> params
            Parameters({'VALUE': 'TEXT'})

        Create new parameters:

        .. code-block:: pycon

            >>> params = Parameters(value="BINARY")
            >>> params.value
            'BINARY'

        Set a default:

        .. code-block:: pycon

            >>> params = Parameters(value="BINARY", default_value="TEXT")
            >>> params
            Parameters({'VALUE': 'BINARY'})

    """

    def __init__(self, *args, **kwargs):
        """Create new parameters."""
        if args and args[0] is None:
            # allow passing None
            args = args[1:]
        defaults = {
            key[8:]: kwargs.pop(key)
            for key in list(kwargs.keys())
            if key.lower().startswith("default_")
        }
        super().__init__(*args, **kwargs)
        for key, value in defaults.items():
            self.setdefault(key, value)

    # The following paremeters must always be enclosed in double quotes
    always_quoted = (
        "ALTREP",
        "DELEGATED-FROM",
        "DELEGATED-TO",
        "DIR",
        "MEMBER",
        "SENT-BY",
        # Part of X-APPLE-STRUCTURED-LOCATION
        "X-ADDRESS",
        "X-TITLE",
        # RFC 9253
        "LINKREL",
    )
    # this is quoted should one of the values be present
    quote_also = {
        # This is escaped in the RFC
        "CN": " '",
    }

    def params(self):
        """In RFC 5545 keys are called parameters, so this is to be consitent
        with the naming conventions.
        """
        return self.keys()

    def to_ical(self, sorted: bool = True):  # noqa: A002, FBT001
        """Returns an :rfc:`5545` representation of the parameters.

        Args:
            sorted (bool): Sort the parameters before encoding.
            exclude_utc (bool): Exclude TZID if it is set to ``"UTC"``
        """
        result = []
        items = list(self.items())
        if sorted:
            items.sort()

        for key, value in items:
            if key == "TZID" and value == "UTC":
                # The "TZID" property parameter MUST NOT be applied to DATE-TIME
                # properties whose time values are specified in UTC.
                continue
            upper_key = key.upper()
            check_quoteable_characters = self.quote_also.get(key.upper())
            always_quote = upper_key in self.always_quoted or (
                check_quoteable_characters
                and any(c in value for c in check_quoteable_characters)
            )
            quoted_value = param_value(value, always_quote=always_quote)
            if isinstance(quoted_value, str):
                quoted_value = quoted_value.encode(DEFAULT_ENCODING)
            # CaselessDict keys are always unicode
            result.append(upper_key.encode(DEFAULT_ENCODING) + b"=" + quoted_value)
        return b";".join(result)

    @classmethod
    def from_ical(cls, st, strict=False):
        """Parses the parameter format from ical text format."""

        # parse into strings
        result = cls()
        for param in q_split(st, ";"):
            try:
                key, val = q_split(param, "=", maxsplit=1)
                validate_token(key)
                # Property parameter values that are not in quoted
                # strings are case insensitive.
                vals = []
                for v in q_split(val, ","):
                    if v.startswith('"') and v.endswith('"'):
                        v2 = v.strip('"')
                        validate_param_value(v2, quoted=True)
                        vals.append(rfc_6868_unescape(v2))
                    else:
                        validate_param_value(v, quoted=False)
                        if strict:
                            vals.append(rfc_6868_unescape(v.upper()))
                        else:
                            vals.append(rfc_6868_unescape(v))
                if not vals:
                    result[key] = val
                elif len(vals) == 1:
                    result[key] = vals[0]
                else:
                    result[key] = vals
            except ValueError as exc:  # noqa: PERF203
                raise ValueError(
                    f"{param!r} is not a valid parameter string: {exc}"
                ) from exc
        return result

    @single_string_parameter(upper=True)
    def value(self) -> VALUE | str | None:
        """The VALUE parameter from :rfc:`5545`.

        Description:
            This parameter specifies the value type and format of
            the property value.  The property values MUST be of a single value
            type.  For example, a "RDATE" property cannot have a combination
            of DATE-TIME and TIME value types.

            If the property's value is the default value type, then this
            parameter need not be specified.  However, if the property's
            default value type is overridden by some other allowable value
            type, then this parameter MUST be specified.

            Applications MUST preserve the value data for x-name and iana-
            token values that they don't recognize without attempting to
            interpret or parse the value data.

        For convenience, using this property, the value will be converted to
        an uppercase string.

        .. code-block:: pycon

            >>> from icalendar import Parameters
            >>> params = Parameters()
            >>> params.value = "unknown"
            >>> params
            Parameters({'VALUE': 'UNKNOWN'})

        """

    def _parameter_value_to_jcal(
        self, value: str | float | list | VPROPERTY
    ) -> str | int | float | list[str] | list[int] | list[float]:
        """Convert a parameter value to jCal format.

        Args:
            value: The parameter value

        Returns:
            The jCal representation of the parameter value
        """
        if isinstance(value, list):
            return [self._parameter_value_to_jcal(v) for v in value]
        if hasattr(value, "to_jcal"):
            # proprty values respond to this
            jcal = value.to_jcal()
            # we only need the value part
            if len(jcal) == 4:
                return jcal[3]
            return jcal[3:]
        for t in (int, float, str):
            if isinstance(value, t):
                return t(value)
        raise TypeError(
            "Unsupported parameter value type for jCal conversion: "
            f"{type(value)} {value!r}"
        )

    def to_jcal(self, exclude_utc=False) -> dict[str, str]:
        """Return the jCal representation of the parameters.

        Args:
            exclude_utc (bool): Exclude the TZID parameter if it is UTC
        """
        jcal = {
            k.lower(): self._parameter_value_to_jcal(v)
            for k, v in self.items()
            if k.lower() != "value"
        }
        if exclude_utc and jcal.get("tzid") == "UTC":
            del jcal["tzid"]
        return jcal

    @single_string_parameter
    def tzid(self) -> str | None:
        """The TZID parameter from :rfc:`5545`."""

    def is_utc(self):
        """Whether the TZID parameter is UTC."""
        return self.tzid == "UTC"

    def update_tzid_from(self, dt: datetime | time | Any) -> None:
        """Update the TZID parameter from a datetime object.

        This sets the TZID parameter or deletes it according to the datetime.
        """
        if isinstance(dt, (datetime, time)):
            self.tzid = tzid_from_dt(dt)

    @classmethod
    def from_jcal(cls, jcal: dict[str : str | list[str]]):
        """Parse jCal parameters."""
        if not isinstance(jcal, dict):
            raise JCalParsingError("The parameters must be a mapping.", cls)
        for name, value in jcal.items():
            if not isinstance(name, str):
                raise JCalParsingError(
                    "All parameter names must be strings.", cls, value=name
                )
            if not (
                (
                    isinstance(value, list)
                    and all(isinstance(v, (str, int, float)) for v in value)
                    and value
                )
                or isinstance(value, (str, int, float))
            ):
                raise JCalParsingError(
                    "Parameter values must be a string, integer or "
                    "float or a list of those.",
                    cls,
                    name,
                    value=value,
                )
        return cls(jcal)

    @classmethod
    def from_jcal_property(cls, jcal_property: list):
        """Create the parameters for a jCal property.

        Args:
            jcal_property (list): The jCal property [name, params, value, ...]
            default_value (str, optional): The default value of the property.
                If this is given, the default value will not be set.
        """
        if not isinstance(jcal_property, list) or len(jcal_property) < 4:
            raise JCalParsingError(
                "The property must be a list with at least 4 items.", cls
            )
        jcal_params = jcal_property[1]
        with JCalParsingError.reraise_with_path_added(1):
            self = cls.from_jcal(jcal_params)
        if self.is_utc():
            del self.tzid  # we do not want this parameter
        return self


def escape_string(val):
    # f'{i:02X}'
    return (
        val.replace(r"\,", "%2C")
        .replace(r"\:", "%3A")
        .replace(r"\;", "%3B")
        .replace(r"\\", "%5C")
    )


def unescape_string(val):
    return (
        val.replace("%2C", ",")
        .replace("%3A", ":")
        .replace("%3B", ";")
        .replace("%5C", "\\")
    )


_unescape_backslash_regex = re.compile(r"\\([\\,;:nN])")


def unescape_backslash(val: str):
    r"""Unescape backslash sequences in iCalendar text.

    Unlike :py:meth:`unescape_string`, this only handles actual backslash escapes
    per :rfc:`5545`, not URL encoding. This preserves URL-encoded values
    like ``%3A`` in URLs.

    Processes backslash escape sequences in a single pass using regex matching.
    """
    return _unescape_backslash_regex.sub(
        lambda m: "\n" if m.group(1) in "nN" else m.group(1), val
    )


RFC_6868_UNESCAPE_REGEX = re.compile(r"\^\^|\^n|\^'")


def rfc_6868_unescape(param_value: str) -> str:
    """Take care of :rfc:`6868` unescaping.

    - ^^ -> ^
    - ^n -> system specific newline
    - ^' -> "
    - ^ with others stay intact
    """
    replacements = {
        "^^": "^",
        "^n": os.linesep,
        "^'": '"',
    }
    return RFC_6868_UNESCAPE_REGEX.sub(
        lambda m: replacements.get(m.group(0), m.group(0)), param_value
    )


RFC_6868_ESCAPE_REGEX = re.compile(r'\^|\r\n|\r|\n|"')


def rfc_6868_escape(param_value: str) -> str:
    """Take care of :rfc:`6868` escaping.

    - ^ -> ^^
    - " -> ^'
    - newline -> ^n
    """
    replacements = {
        "^": "^^",
        "\n": "^n",
        "\r": "^n",
        "\r\n": "^n",
        '"': "^'",
    }
    return RFC_6868_ESCAPE_REGEX.sub(
        lambda m: replacements.get(m.group(0), m.group(0)), param_value
    )


def unescape_list_or_string(val):
    if isinstance(val, list):
        return [unescape_string(s) for s in val]
    return unescape_string(val)


#########################################
# parsing and generation of content lines


class Contentline(str):
    """A content line is basically a string that can be folded and parsed into
    parts.
    """

    __slots__ = ("strict",)

    def __new__(cls, value, strict=False, encoding=DEFAULT_ENCODING):
        value = to_unicode(value, encoding=encoding)
        assert "\n" not in value, (
            "Content line can not contain unescaped new line characters."
        )
        self = super().__new__(cls, value)
        self.strict = strict
        return self

    @classmethod
    def from_parts(
        cls,
        name: ICAL_TYPE,
        params: Parameters,
        values,
        sorted: bool = True,  # noqa: A002, FBT001
    ):
        """Turn a parts into a content line."""
        assert isinstance(params, Parameters)
        if hasattr(values, "to_ical"):
            values = values.to_ical()
        else:
            from icalendar.prop import vText

            values = vText(values).to_ical()
        # elif isinstance(values, basestring):
        #    values = escape_char(values)

        # TODO: after unicode only, remove this
        # Convert back to unicode, after to_ical encoded it.
        name = to_unicode(name)
        values = to_unicode(values)
        if params:
            params = to_unicode(params.to_ical(sorted=sorted))
            if params:
                # some parameter values can be skipped during serialization
                return cls(f"{name};{params}:{values}")
        return cls(f"{name}:{values}")

    def parts(self) -> tuple[str, Parameters, str]:
        """Split the content line into ``name``, ``parameters``, and ``values`` parts.

        Properly handles escaping with backslashes and double-quote sections
        to avoid corrupting URL-encoded characters in values.

        Example with parameter:

        .. code-block:: text

            DESCRIPTION;ALTREP="cid:part1.0001@example.org":The Fall'98 Wild

        Example without parameters:

        .. code-block:: text

            DESCRIPTION:The Fall'98 Wild
        """
        try:
            name_split: int | None = None
            value_split: int | None = None
            in_quotes: bool = False
            escaped: bool = False

            for i, ch in enumerate(self):
                if ch == '"' and not escaped:
                    in_quotes = not in_quotes
                elif ch == "\\" and not in_quotes:
                    escaped = True
                    continue
                elif not in_quotes and not escaped:
                    # Find first delimiter for name
                    if ch in ":;" and name_split is None:
                        name_split = i
                    # Find value delimiter (first colon)
                    if ch == ":" and value_split is None:
                        value_split = i

                escaped = False

            # Validate parsing results
            if not value_split:
                # No colon found - value is empty, use end of string
                value_split = len(self)

            # Extract name - if no delimiter,
            #   take whole string for validate_token to reject
            name = self[:name_split] if name_split else self
            validate_token(name)

            if not name_split or name_split + 1 == value_split:
                # No delimiter or empty parameter section
                raise ValueError("Invalid content line")  # noqa: TRY301
            # Parse parameters - they still need to be escaped/unescaped
            # for proper handling of commas, semicolons, etc. in parameter values
            param_str = escape_string(self[name_split + 1 : value_split])
            params = Parameters.from_ical(param_str, strict=self.strict)
            params = Parameters(
                (unescape_string(key), unescape_list_or_string(value))
                for key, value in iter(params.items())
            )
            # Unescape backslash sequences in values but preserve URL encoding
            values = unescape_backslash(self[value_split + 1 :])
        except ValueError as exc:
            raise ValueError(
                f"Content line could not be parsed into parts: '{self}': {exc}"
            ) from exc
        return (name, params, values)

    @classmethod
    def from_ical(cls, ical, strict=False):
        """Unfold the content lines in an iCalendar into long content lines."""
        ical = to_unicode(ical)
        # a fold is carriage return followed by either a space or a tab
        return cls(UFOLD.sub("", ical), strict=strict)

    def to_ical(self):
        """Long content lines are folded so they are less than 75 characters
        wide.
        """
        return foldline(self).encode(DEFAULT_ENCODING)


class Contentlines(list):
    """I assume that iCalendar files generally are a few kilobytes in size.
    Then this should be efficient. for Huge files, an iterator should probably
    be used instead.
    """

    def to_ical(self):
        """Simply join self."""
        return b"\r\n".join(line.to_ical() for line in self if line) + b"\r\n"

    @classmethod
    def from_ical(cls, st):
        """Parses a string into content lines."""
        st = to_unicode(st)
        try:
            # a fold is carriage return followed by either a space or a tab
            unfolded = UFOLD.sub("", st)
            lines = cls(Contentline(line) for line in NEWLINE.split(unfolded) if line)
            lines.append("")  # '\r\n' at the end of every content line
        except Exception as e:
            raise ValueError("Expected StringType with content lines") from e
        return lines


__all__ = [
    "FOLD",
    "NAME",
    "NEWLINE",
    "QUNSAFE_CHAR",
    "QUOTABLE",
    "UFOLD",
    "UNSAFE_CHAR",
    "Contentline",
    "Contentlines",
    "Parameters",
    "dquote",
    "escape_char",
    "escape_string",
    "foldline",
    "param_value",
    "q_join",
    "q_split",
    "rfc_6868_escape",
    "rfc_6868_unescape",
    "unescape_backslash",
    "unescape_char",
    "unescape_list_or_string",
    "unescape_string",
    "validate_param_value",
    "validate_token",
]



"""Test the properties of the alarm."""

import pytest

from icalendar.cal.alarm import Alarm
from icalendar.error import InvalidCalendar


def test_repeat_absent():
    """Test the absence of REPEAT."""
    assert Alarm().REPEAT == 0


def test_repeat_number():
    """Test the absence of REPEAT."""
    assert Alarm({"REPEAT": 10}).REPEAT == 10


def test_set_REPEAT():
    """Check setting the value."""
    a = Alarm()
    a.REPEAT = 10
    assert a.REPEAT == 10


def test_set_REPEAT_twice():
    """Check setting the value."""
    a = Alarm()
    a.REPEAT = 10
    a.REPEAT = 20
    assert a.REPEAT == 20


def test_add_REPEAT():
    """Check setting the value."""
    a = Alarm()
    a.add("REPEAT", 10)
    assert a.REPEAT == 10


def test_invalid_repeat_value():
    """Check setting the value."""
    a = Alarm()
    with pytest.raises(ValueError):
        a.REPEAT = "asd"
    a["REPEAT"] = "asd"
    with pytest.raises(InvalidCalendar):
        a.REPEAT  # noqa: B018


def test_alarm_to_string():
    a = Alarm()
    a.REPEAT = 11
    assert a.to_ical() == b"BEGIN:VALARM\r\nREPEAT:11\r\nEND:VALARM\r\n"



import unittest
from datetime import date, datetime, time, timedelta

from icalendar.parser import Parameters


class TestProp(unittest.TestCase):
    def test_prop_vFloat(self):
        from icalendar.prop import vFloat

        assert vFloat(1.0).to_ical() == b"1.0"
        assert vFloat.from_ical("42") == 42.0
        assert vFloat(42).to_ical() == b"42.0"
        self.assertRaises(ValueError, vFloat.from_ical, "1s3")

    def test_prop_vInt(self):
        from icalendar.prop import vInt

        assert vInt(42).to_ical() == b"42"
        assert vInt.from_ical("13") == 13
        self.assertRaises(ValueError, vInt.from_ical, "1s3")

    def test_prop_vDDDLists(self):
        from icalendar.prop import vDDDLists

        dt_list = vDDDLists.from_ical("19960402T010000Z")
        assert isinstance(dt_list, list)
        assert len(dt_list) == 1
        assert isinstance(dt_list[0], datetime)
        assert str(dt_list[0]) == "1996-04-02 01:00:00+00:00"

        p = "19960402T010000Z,19960403T010000Z,19960404T010000Z"
        dt_list = vDDDLists.from_ical(p)
        assert len(dt_list) == 3
        assert str(dt_list[0]) == "1996-04-02 01:00:00+00:00"
        assert str(dt_list[2]) == "1996-04-04 01:00:00+00:00"

        dt_list = vDDDLists([])
        assert dt_list.to_ical() == b""

        dt_list = vDDDLists([datetime(2000, 1, 1)])
        assert dt_list.to_ical() == b"20000101T000000"

        dt_list = vDDDLists([datetime(2000, 1, 1), datetime(2000, 11, 11)])
        assert dt_list.to_ical() == b"20000101T000000,20001111T000000"

        instance = vDDDLists([])
        assert instance != "value"

    def test_prop_vDate(self):
        from icalendar.prop import vDate

        assert vDate(date(2001, 1, 1)).to_ical() == b"20010101"
        assert vDate(date(1899, 1, 1)).to_ical() == b"18990101"

        assert vDate.from_ical("20010102") == date(2001, 1, 2)

        self.assertRaises(TypeError, vDate, "d")
        self.assertRaises(ValueError, vDate.from_ical, "200102")

    def test_prop_vDuration(self):
        from icalendar.prop import vDuration

        assert vDuration(timedelta(11)).to_ical() == b"P11D"
        assert vDuration(timedelta(-14)).to_ical() == b"-P14D"
        assert vDuration(timedelta(1, 7384)).to_ical() == b"P1DT2H3M4S"
        assert vDuration(timedelta(1, 7380)).to_ical() == b"P1DT2H3M"
        assert vDuration(timedelta(1, 7200)).to_ical() == b"P1DT2H"
        assert vDuration(timedelta(0, 7200)).to_ical() == b"PT2H"
        assert vDuration(timedelta(0, 7384)).to_ical() == b"PT2H3M4S"
        assert vDuration(timedelta(0, 184)).to_ical() == b"PT3M4S"
        assert vDuration(timedelta(0, 22)).to_ical() == b"PT22S"
        assert vDuration(timedelta(0, 3622)).to_ical() == b"PT1H0M22S"
        assert vDuration(timedelta(days=1, hours=5)).to_ical() == b"P1DT5H"
        assert vDuration(timedelta(hours=-5)).to_ical() == b"-PT5H"
        assert vDuration(timedelta(days=-1, hours=-5)).to_ical() == b"-P1DT5H"

        # How does the parsing work?
        assert vDuration.from_ical("PT1H0M22S") == timedelta(0, 3622)

        self.assertRaises(ValueError, vDuration.from_ical, "kox")

        assert vDuration.from_ical("-P14D") == timedelta(-14)

        self.assertRaises(TypeError, vDuration, 11)

        # calling to_ical twice should result in same output
        duration = vDuration(timedelta(days=-1, hours=-5))
        assert duration.to_ical() == b"-P1DT5H"
        assert duration.to_ical() == b"-P1DT5H"

    def test_prop_vWeekday(self):
        from icalendar.prop import vWeekday

        assert vWeekday("mo").to_ical() == b"MO"
        self.assertRaises(ValueError, vWeekday, "erwer")
        assert vWeekday.from_ical("mo") == "MO"
        assert vWeekday.from_ical("+3mo") == "+3MO"
        self.assertRaises(ValueError, vWeekday.from_ical, "Saturday")
        assert vWeekday("+mo").to_ical() == b"+MO"
        assert vWeekday("+3mo").to_ical() == b"+3MO"
        assert vWeekday("-tu").to_ical() == b"-TU"

    def test_prop_vFrequency(self):
        from icalendar.prop import vFrequency

        self.assertRaises(ValueError, vFrequency, "bad test")
        assert vFrequency("daily").to_ical() == b"DAILY"
        assert vFrequency("daily").from_ical("MONTHLY") == "MONTHLY"
        self.assertRaises(ValueError, vFrequency.from_ical, 234)

    def test_prop_vRecur(self):
        from icalendar.prop import vRecur

        # Let's see how close we can get to one from the rfc:
        # FREQ=YEARLY;INTERVAL=2;BYMONTH=1;BYDAY=SU;BYHOUR=8,9;BYMINUTE=30

        r = {"freq": "yearly", "interval": 2}
        r.update({"bymonth": 1, "byday": "su", "byhour": [8, 9], "byminute": 30})
        assert (
            vRecur(r).to_ical()
            == b"FREQ=YEARLY;INTERVAL=2;BYMINUTE=30;BYHOUR=8,9;BYDAY=SU;BYMONTH=1"
        )

        r = vRecur(FREQ="yearly", INTERVAL=2)
        r.update(
            {
                "BYMONTH": 1,
                "BYDAY": "su",
                "BYHOUR": [8, 9],
                "BYMINUTE": 30,
            }
        )
        assert (
            r.to_ical()
            == b"FREQ=YEARLY;INTERVAL=2;BYMINUTE=30;BYHOUR=8,9;BYDAY=SU;BYMONTH=1"
        )

        r = vRecur(freq="DAILY", count=10)
        r["bysecond"] = [0, 15, 30, 45]
        assert r.to_ical() == b"FREQ=DAILY;COUNT=10;BYSECOND=0,15,30,45"

        r = vRecur(freq="DAILY", until=datetime(2005, 1, 1, 12, 0, 0))
        assert r.to_ical() == b"FREQ=DAILY;UNTIL=20050101T120000"

        # How do we fare with regards to parsing?
        r = vRecur.from_ical("FREQ=DAILY;INTERVAL=2;COUNT=10")
        assert r == {"COUNT": [10], "FREQ": ["DAILY"], "INTERVAL": [2]}
        assert vRecur(r).to_ical() == b"FREQ=DAILY;COUNT=10;INTERVAL=2"

        r = vRecur.from_ical(
            "FREQ=YEARLY;INTERVAL=2;BYMONTH=1;BYDAY=-SU;BYHOUR=8,9;BYMINUTE=30"
        )
        assert r == {
            "BYHOUR": [8, 9],
            "BYDAY": ["-SU"],
            "BYMINUTE": [30],
            "BYMONTH": [1],
            "FREQ": ["YEARLY"],
            "INTERVAL": [2],
        }

        assert (
            vRecur(r).to_ical()
            == b"FREQ=YEARLY;INTERVAL=2;BYMINUTE=30;BYHOUR=8,9;BYDAY=-SU;BYMONTH=1"
        )

        r = vRecur.from_ical("FREQ=WEEKLY;INTERVAL=1;BYWEEKDAY=TH")

        assert r == {"FREQ": ["WEEKLY"], "INTERVAL": [1], "BYWEEKDAY": ["TH"]}

        assert vRecur(r).to_ical() == b"FREQ=WEEKLY;INTERVAL=1;BYWEEKDAY=TH"

        # Some examples from the spec
        r = vRecur.from_ical("FREQ=MONTHLY;BYDAY=MO,TU,WE,TH,FR;BYSETPOS=-1")
        assert vRecur(r).to_ical() == b"FREQ=MONTHLY;BYDAY=MO,TU,WE,TH,FR;BYSETPOS=-1"

        p = "FREQ=YEARLY;INTERVAL=2;BYMONTH=1;BYDAY=SU;BYHOUR=8,9;BYMINUTE=30"
        r = vRecur.from_ical(p)
        assert (
            vRecur(r).to_ical()
            == b"FREQ=YEARLY;INTERVAL=2;BYMINUTE=30;BYHOUR=8,9;BYDAY=SU;BYMONTH=1"
        )

        # and some errors
        self.assertRaises(ValueError, vRecur.from_ical, "BYDAY=12")

        # when key is not RFC-compliant, parse it as vText
        r = vRecur.from_ical("FREQ=MONTHLY;BYOTHER=TEXT;BYEASTER=-3")
        assert vRecur(r).to_ical() == b"FREQ=MONTHLY;BYEASTER=-3;BYOTHER=TEXT"

    def test_prop_vText(self):
        from icalendar.prop import vText

        assert vText("Simple text").to_ical() == b"Simple text"

        # Escaped text
        t = vText("Text ; with escaped, chars")
        assert t.to_ical() == b"Text \\; with escaped\\, chars"

        # Escaped newlines
        assert (
            vText("Text with escaped\\N chars").to_ical()
            == b"Text with escaped\\n chars"
        )

        # If you pass a unicode object, it will be utf-8 encoded. As this is
        # the (only) standard that RFC 5545 support.
        t = vText("international chars \xe4\xf6\xfc")
        assert t.to_ical() == b"international chars \xc3\xa4\xc3\xb6\xc3\xbc"

        # and parsing?
        assert (
            vText.from_ical("Text \\; with escaped\\, chars")
            == "Text ; with escaped, chars"
        )

        t = vText.from_ical("A string with\\; some\\\\ characters in\\it")
        assert t == "A string with; some\\ characters in\\it"

        # We are forgiving to utf-8 encoding errors:
        # We intentionally use a string with unexpected encoding
        #
        assert vText.from_ical(b"Ol\xe9") == "Ol�"

        # Notice how accented E character, encoded with latin-1, got replaced
        # with the official U+FFFD REPLACEMENT CHARACTER.

    def test_prop_vTime(self):
        from icalendar.prop import vTime

        assert vTime(12, 30, 0).to_ical() == "123000"
        assert vTime.from_ical("123000") == time(12, 30)

        # We should also fail, right?
        self.assertRaises(ValueError, vTime.from_ical, "263000")

        self.assertRaises(ValueError, vTime, "263000")

    def test_prop_vUri(self):
        from icalendar.prop import vUri

        assert vUri("http://www.example.com/").to_ical() == b"http://www.example.com/"
        assert vUri.from_ical("http://www.example.com/") == "http://www.example.com/"

    def test_prop_vGeo(self):
        from icalendar.prop import vGeo

        # Pass a list
        assert vGeo([1.2, 3.0]).to_ical() == "1.2;3.0"

        # Pass a tuple
        assert vGeo((1.2, 3.0)).to_ical() == "1.2;3.0"

        g = vGeo.from_ical("37.386013;-122.082932")
        assert g == (float("37.386013"), float("-122.082932"))

        assert vGeo(g).to_ical() == "37.386013;-122.082932"

        self.assertRaises(ValueError, vGeo, "g")
        self.assertRaises(ValueError, vGeo.from_ical, "1s3;1s3")

    def test_prop_vUTCOffset(self):
        from icalendar.prop import vUTCOffset

        assert vUTCOffset(timedelta(hours=2)).to_ical() == "+0200"

        assert vUTCOffset(timedelta(hours=-5)).to_ical() == "-0500"

        assert vUTCOffset(timedelta()).to_ical() == "+0000"

        assert vUTCOffset(timedelta(minutes=-30)).to_ical() == "-0030"

        assert vUTCOffset(timedelta(hours=2, minutes=-30)).to_ical() == "+0130"

        assert vUTCOffset(timedelta(hours=1, minutes=30)).to_ical() == "+0130"

        # Support seconds
        assert (
            vUTCOffset(timedelta(hours=1, minutes=30, seconds=7)).to_ical() == "+013007"
        )

        # Parsing

        assert vUTCOffset.from_ical("0000") == timedelta(0)
        assert vUTCOffset.from_ical("-0030") == timedelta(-1, 84600)
        assert vUTCOffset.from_ical("+0200") == timedelta(0, 7200)
        assert vUTCOffset.from_ical("+023040") == timedelta(0, 9040)

        assert vUTCOffset(vUTCOffset.from_ical("+0230")).to_ical() == "+0230"

        # And a few failures
        self.assertRaises(ValueError, vUTCOffset.from_ical, "+323k")

        self.assertRaises(ValueError, vUTCOffset.from_ical, "+2400")

        self.assertRaises(TypeError, vUTCOffset, "0:00:00")

    def test_prop_vInline(self):
        from icalendar.prop import vInline

        assert vInline("Some text") == "Some text"
        assert vInline("Some text").to_ical() == b"Some text"
        assert vInline.from_ical("Some text") == "Some text"

        t2 = vInline("other text")
        t2.params["cn"] = "Test Osterone"
        assert isinstance(t2.params, Parameters)
        assert t2.params == {"CN": "Test Osterone"}

    def test_prop_vCategory(self):
        from icalendar.prop import vCategory

        catz = ["cat 1", "cat 2", "cat 3"]
        v_cat = vCategory(catz)

        assert v_cat.to_ical() == b"cat 1,cat 2,cat 3"
        assert vCategory.from_ical(v_cat.to_ical()) == catz
        c = vCategory(vCategory.from_ical("APPOINTMENT,EDUCATION"))
        cats = list(c)
        assert cats == ["APPOINTMENT", "EDUCATION"]

    def test_prop_TypesFactory(self):
        from icalendar.prop import TypesFactory

        # To get a type you can use it like this.
        factory = TypesFactory()
        datetime_parser = factory["date-time"]
        assert datetime_parser(datetime(2001, 1, 1)).to_ical() == b"20010101T000000"

        # A typical use is when the parser tries to find a content type and use
        # text as the default
        value = "20050101T123000"
        value_type = "date-time"
        assert factory.get(value_type, "text").from_ical(value) == datetime(
            2005, 1, 1, 12, 30
        )

        # It can also be used to directly encode property and parameter values
        assert (
            factory.to_ical("comment", "by Rasmussen, Max Müller")
            == b"by Rasmussen\\, Max M\xc3\xbcller"
        )
        assert factory.to_ical("priority", 1) == b"1"
        assert (
            factory.to_ical("cn", "Rasmussen, Max Müller")
            == b"Rasmussen\\, Max M\xc3\xbcller"
        )
        assert (
            factory.from_ical("cn", b"Rasmussen\\, Max M\xc3\xb8ller")
            == "Rasmussen, Max Møller"
        )



import unittest

import icalendar


class TestCaselessdict(unittest.TestCase):
    def test_caselessdict_canonsort_keys(self):
        canonsort_keys = icalendar.caselessdict.canonsort_keys

        keys = ["DTEND", "DTSTAMP", "DTSTART", "UID", "SUMMARY", "LOCATION"]

        out = canonsort_keys(keys)
        self.assertEqual(
            out, ["DTEND", "DTSTAMP", "DTSTART", "LOCATION", "SUMMARY", "UID"]
        )

        out = canonsort_keys(
            keys,
            (
                "SUMMARY",
                "DTSTART",
                "DTEND",
            ),
        )
        self.assertEqual(
            out, ["SUMMARY", "DTSTART", "DTEND", "DTSTAMP", "LOCATION", "UID"]
        )

        out = canonsort_keys(
            keys,
            (
                "UID",
                "DTSTART",
                "DTEND",
            ),
        )
        self.assertEqual(
            out, ["UID", "DTSTART", "DTEND", "DTSTAMP", "LOCATION", "SUMMARY"]
        )

        out = canonsort_keys(keys, ("UID", "DTSTART", "DTEND", "RRULE", "EXDATE"))
        self.assertEqual(
            out, ["UID", "DTSTART", "DTEND", "DTSTAMP", "LOCATION", "SUMMARY"]
        )

    def test_caselessdict_canonsort_items(self):
        canonsort_items = icalendar.caselessdict.canonsort_items

        d = {
            "i": 7,
            "c": "at",
            "a": 3.5,
            "l": (2, 3),
            "e": [4, 5],
            "n": 13,
            "d": {"x": "y"},
            "r": 1.0,
        }

        out = canonsort_items(d)
        self.assertEqual(
            out,
            [
                ("a", 3.5),
                ("c", "at"),
                ("d", {"x": "y"}),
                ("e", [4, 5]),
                ("i", 7),
                ("l", (2, 3)),
                ("n", 13),
                ("r", 1.0),
            ],
        )

        out = canonsort_items(d, ("i", "c", "a"))
        self.assertTrue(
            out,
            [
                ("i", 7),
                ("c", "at"),
                ("a", 3.5),
                ("d", {"x": "y"}),
                ("e", [4, 5]),
                ("l", (2, 3)),
                ("n", 13),
                ("r", 1.0),
            ],
        )

    def test_caselessdict_copy(self):
        CaselessDict = icalendar.caselessdict.CaselessDict

        original_dict = CaselessDict(key1="val1", key2="val2")
        copied_dict = original_dict.copy()

        self.assertEqual(original_dict, copied_dict)

    def test_CaselessDict(self):
        CaselessDict = icalendar.caselessdict.CaselessDict

        ncd = CaselessDict(key1="val1", key2="val2")
        self.assertEqual(ncd, CaselessDict({"KEY2": "val2", "KEY1": "val1"}))

        self.assertEqual(ncd["key1"], "val1")
        self.assertEqual(ncd["KEY1"], "val1")

        assert ncd.has_key("key1")

        ncd["KEY3"] = "val3"
        self.assertEqual(ncd["key3"], "val3")

        self.assertEqual(ncd.setdefault("key3", "FOUND"), "val3")
        self.assertEqual(ncd.setdefault("key4", "NOT FOUND"), "NOT FOUND")
        self.assertEqual(ncd["key4"], "NOT FOUND")
        self.assertEqual(ncd.get("key1"), "val1")
        self.assertEqual(ncd.get("key3", "NOT FOUND"), "val3")
        self.assertEqual(ncd.get("key4", "NOT FOUND"), "NOT FOUND")
        self.assertIn("key4", ncd)

        del ncd["key4"]
        self.assertNotIn("key4", ncd)

        ncd.update({"key5": "val5", "KEY6": "val6", "KEY5": "val7"})
        self.assertEqual(ncd["key6"], "val6")

        keys = sorted(ncd.keys())
        self.assertEqual(keys, ["KEY1", "KEY2", "KEY3", "KEY5", "KEY6"])



"""Tests for issue #870: Add new() method to vCalAddress class."""

from datetime import datetime

import pytest

from icalendar import Event
from icalendar.prop import vCalAddress


class TestVCalAddressNew:
    """Test the vCalAddress.new() method functionality."""

    def test_new_with_email_only(self):
        """Test creating vCalAddress with email only."""
        addr = vCalAddress.new("test@test.com")

        assert str(addr) == "mailto:test@test.com"
        assert len(addr.params) == 0

    def test_new_with_mailto_prefix(self):
        """Test that existing mailto: prefix is preserved."""
        addr = vCalAddress.new("mailto:test@test.com")

        assert str(addr) == "mailto:test@test.com"

    def test_new_with_case_insensitive_mailto(self):
        """Test case-insensitive mailto: handling."""
        test_cases = [
            "MAILTO:test@test.com",
            "Mailto:test@test.com",
            "MailTo:test@test.com",
            "mAiLtO:test@test.com",
        ]

        for email in test_cases:
            addr = vCalAddress.new(email)
            assert str(addr) == email  # Preserve original case

    def test_new_with_cn_parameter(self):
        """Test creating vCalAddress with CN parameter."""
        addr = vCalAddress.new("test@test.com", cn="Test User")

        assert str(addr) == "mailto:test@test.com"
        assert addr.params["CN"] == "Test User"

    @pytest.mark.parametrize(
        ("argument", "value", "parameter", "expected_value"),
        [
            ("cutype", "INDIVIDUAL", "CUTYPE", "INDIVIDUAL"),
            ("cutype", "GROUP", "CUTYPE", "GROUP"),
            ("cutype", "RESOURCE", "CUTYPE", "RESOURCE"),
            ("cutype", "ROOM", "CUTYPE", "ROOM"),
            ("role", "REQ-PARTICIPANT", "ROLE", "REQ-PARTICIPANT"),
            ("role", "OPT-PARTICIPANT", "ROLE", "OPT-PARTICIPANT"),
            ("role", "NON-PARTICIPANT", "ROLE", "NON-PARTICIPANT"),
            ("role", "CHAIR", "ROLE", "CHAIR"),
            ("partstat", "NEEDS-ACTION", "PARTSTAT", "NEEDS-ACTION"),
            ("partstat", "ACCEPTED", "PARTSTAT", "ACCEPTED"),
            ("partstat", "DECLINED", "PARTSTAT", "DECLINED"),
            ("partstat", "TENTATIVE", "PARTSTAT", "TENTATIVE"),
            ("partstat", "DELEGATED", "PARTSTAT", "DELEGATED"),
            ("language", "en-US", "LANGUAGE", "en-US"),
            ("language", "de-DE", "LANGUAGE", "de-DE"),
            ("directory", "ldap://example.com", "DIR", "ldap://example.com"),
        ],
    )
    def test_new_with_string_parameters(
        self, argument, value, parameter, expected_value
    ):
        """Test vCalAddress.new() with various string parameters."""
        kwargs = {argument: value}
        addr = vCalAddress.new("test@test.com", **kwargs)

        assert str(addr) == "mailto:test@test.com"
        assert addr.params[parameter] == expected_value

    @pytest.mark.parametrize(
        ("rsvp_value", "expected_param"),
        [
            (True, "TRUE"),
            (False, "FALSE"),
        ],
    )
    def test_new_with_rsvp_parameter(self, rsvp_value, expected_param):
        """Test vCalAddress.new() with RSVP parameter."""
        addr = vCalAddress.new("test@test.com", rsvp=rsvp_value)

        assert str(addr) == "mailto:test@test.com"
        assert addr.params["RSVP"] == expected_param

    @pytest.mark.parametrize(
        ("delegate_param", "delegate_value", "expected_param", "expected_value"),
        [
            (
                "delegated_from",
                "sender@test.com",
                "DELEGATED-FROM",
                "mailto:sender@test.com",
            ),
            (
                "delegated_from",
                "mailto:sender@test.com",
                "DELEGATED-FROM",
                "mailto:sender@test.com",
            ),
            (
                "delegated_to",
                "delegate@test.com",
                "DELEGATED-TO",
                "mailto:delegate@test.com",
            ),
            (
                "delegated_to",
                "MAILTO:delegate@test.com",
                "DELEGATED-TO",
                "MAILTO:delegate@test.com",
            ),
            ("sent_by", "secretary@test.com", "SENT-BY", "mailto:secretary@test.com"),
            (
                "sent_by",
                "mailto:secretary@test.com",
                "SENT-BY",
                "mailto:secretary@test.com",
            ),
        ],
    )
    def test_new_with_email_parameters(
        self, delegate_param, delegate_value, expected_param, expected_value
    ):
        """Test vCalAddress.new() with email delegation parameters."""
        kwargs = {delegate_param: delegate_value}
        addr = vCalAddress.new("test@test.com", **kwargs)

        assert str(addr) == "mailto:test@test.com"
        assert addr.params[expected_param] == expected_value

    def test_new_with_all_parameters(self):
        """Test creating vCalAddress with all parameters."""
        addr = vCalAddress.new(
            "test@test.com",
            cn="Test User",
            cutype="INDIVIDUAL",
            delegated_from="sender@test.com",
            delegated_to="delegate@test.com",
            directory="ldap://example.com",
            language="en-US",
            partstat="ACCEPTED",
            role="REQ-PARTICIPANT",
            rsvp=True,
            sent_by="secretary@test.com",
        )

        assert str(addr) == "mailto:test@test.com"
        assert addr.params["CN"] == "Test User"
        assert addr.params["CUTYPE"] == "INDIVIDUAL"
        assert addr.params["DELEGATED-FROM"] == "mailto:sender@test.com"
        assert addr.params["DELEGATED-TO"] == "mailto:delegate@test.com"
        assert addr.params["DIR"] == "ldap://example.com"
        assert addr.params["LANGUAGE"] == "en-US"
        assert addr.params["PARTSTAT"] == "ACCEPTED"
        assert addr.params["ROLE"] == "REQ-PARTICIPANT"
        assert addr.params["RSVP"] == "TRUE"
        assert addr.params["SENT-BY"] == "mailto:secretary@test.com"

    def test_new_with_none_parameters_ignored(self):
        """Test that None parameters are ignored."""
        addr = vCalAddress.new(
            "test@test.com",
            cn=None,
            role=None,
            rsvp=None,
        )

        assert str(addr) == "mailto:test@test.com"
        assert "CN" not in addr.params
        assert "ROLE" not in addr.params
        assert "RSVP" not in addr.params

    def test_new_integration_with_event(self):
        """Test that vCalAddress.new() works with Event attendees."""
        addr = vCalAddress.new("test@test.com", cn="Test User", rsvp=True)
        event = Event.new(attendees=[addr])

        assert event.attendees == [addr]
        attendee = event.attendees[0]
        assert str(attendee) == "mailto:test@test.com"
        assert attendee.params["CN"] == "Test User"
        assert attendee.params["RSVP"] == "TRUE"


class TestVCalAddressNewErrorCases:
    """Test error cases for vCalAddress.new() method."""

    def test_new_requires_email(self):
        """Test that email parameter is required."""
        with pytest.raises(TypeError):
            vCalAddress.new()  # Missing required email argument

    def test_new_email_must_be_string(self):
        """Test that email must be a string."""
        with pytest.raises(TypeError, match="Email must be a string, not int"):
            vCalAddress.new(123)

        with pytest.raises(TypeError, match="Email must be a string, not list"):
            vCalAddress.new(["test@test.com"])


class TestVCalAddressNewExamples:
    """Test examples that demonstrate vCalAddress.new() usage."""

    def test_basic_usage_example(self):
        """Test basic usage example from docstring."""
        addr = vCalAddress.new("test@test.com")
        assert str(addr) == "mailto:test@test.com"

    def test_with_parameters_example(self):
        """Test example with parameters from docstring."""
        addr = vCalAddress.new("test@test.com", cn="Test User", role="CHAIR")
        assert addr.params["CN"] == "Test User"
        assert addr.params["ROLE"] == "CHAIR"

    def test_consistent_with_event_new(self):
        """Test that vCalAddress.new() integrates well with Event.new()."""
        # Both should create objects with .new() class method
        event = Event.new(summary="Test", start=datetime(2026, 1, 1, 12, 0))
        addr = vCalAddress.new("test@test.com", cn="Test User")

        # Should be able to add the address to event attendees
        event.attendees = [addr]
        assert len(event.attendees) == 1
        assert event.attendees[0].params["CN"] == "Test User"



"""Test vBinary"""

import pytest

from icalendar import vBinary
from icalendar.parser import Parameters


def test_text():
    txt = b"This is gibberish"
    txt_ical = b"VGhpcyBpcyBnaWJiZXJpc2g="
    assert vBinary(txt).to_ical() == txt_ical
    assert vBinary.from_ical(txt_ical) == txt


def test_binary():
    txt = b"Binary data \x13 \x56"
    txt_ical = b"QmluYXJ5IGRhdGEgEyBW"
    assert vBinary(txt).to_ical() == txt_ical
    assert vBinary.from_ical(txt_ical) == txt


def test_param():
    assert isinstance(vBinary("txt").params, Parameters)
    assert vBinary("txt").params == {"VALUE": "BINARY", "ENCODING": "BASE64"}


def test_long_data():
    """Long data should not have line breaks, as that would interfere"""
    txt = b"a" * 99
    txt_ical = b"YWFh" * 33
    assert vBinary(txt).to_ical() == txt_ical
    assert vBinary.from_ical(txt_ical) == txt


def test_repr():
    instance = vBinary("value")
    assert repr(instance) == "vBinary(b'dmFsdWU=')"


def test_from_ical():
    with pytest.raises(ValueError, match="Not valid base 64 encoding."):
        vBinary.from_ical("value")
    with pytest.raises(ValueError, match="Not valid base 64 encoding."):
        vBinary.from_ical("áèਮ")



"""These are tests for considerations that are specific to our implementation."""

import pytest

from icalendar import Calendar, Component, ComponentFactory, Event, vText


def test_adding_a_description_several_times_works():
    """We should be able to add valid parameters several times."""
    event = Event()
    en = vText("English description", params={"language": "en"})
    de = vText("Deutsche Beschreibung", params={"language": "de"})
    event.add("DESCRIPTION", en)
    event.add("DESCRIPTION", de)
    jcal = event.to_jcal()
    assert jcal[1][0] == [
        "description",
        {"language": "en"},
        "text",
        "English description",
    ]
    assert jcal[1][1] == [
        "description",
        {"language": "de"},
        "text",
        "Deutsche Beschreibung",
    ]


def test_get_new_component_class():
    """A new component class should be created."""
    factory = ComponentFactory()
    cls1 = factory.get_component_class("VUNKNOWN")
    cls2 = factory.get_component_class("VUNKNOWN")
    assert cls1 is cls2
    assert cls1.name == "VUNKNOWN"
    assert cls1 is not Component


def test_text_from_jcal_does_not_add_backslashes():
    """It is possible that backslashes are added.

    see https://github.com/collective/icalendar/pull/979#issuecomment-3568261726
    """
    v_text = vText.from_jcal(["name", {}, "unknown", "8 backslash \\\\\\\\"])
    assert v_text.ical_value == "8 backslash \\\\\\\\"


CALENDAR_NAME = "8 backslash \\\\\\\\"


@pytest.fixture
def calendar_with_name():
    """The calendar for escape tests."""
    cal = Calendar()
    cal.calendar_name = CALENDAR_NAME
    assert cal.calendar_name == CALENDAR_NAME, "Setter ad Getter work."
    return cal


def test_backslash_escape_ics(calendar_with_name):
    """It is somehow possible that more backslashes appear.

    See https://github.com/collective/icalendar/issues/1008
    """
    pytest.skip("TODO: Fix in https://github.com/collective/icalendar/issues/1008.")
    ics = calendar_with_name.to_ical()
    cal2: Calendar = Calendar.from_ical(ics)
    assert CALENDAR_NAME == cal2.calendar_name, "ics works"


def test_backslash_escape_jCal(calendar_with_name):
    """It is somehow possible that more backslashes appear.

    See https://github.com/collective/icalendar/issues/1008
    """
    jcal = calendar_with_name.to_jcal()
    cal3: Calendar = Calendar.from_jcal(jcal)
    assert cal3.calendar_name == CALENDAR_NAME, "jcal works"



#!/usr/bin/env python3
"""utility program that allows user to preview calendar's events"""

import argparse
import sys
from datetime import datetime

from icalendar import __version__
from icalendar.cal.calendar import Calendar


def _format_name(address):
    """Retrieve the e-mail and the name from an address.

    :arg an address object, e.g. mailto:test@test.test

    :returns str: The name and the e-mail address.
    """
    email = address.split(":")[-1]
    name = email.split("@")[0]
    if not email:
        return ""
    return f"{name} <{email}>"


def _format_attendees(attendees):
    """Format the list of attendees.

    :arg any attendees: Either a list, a string or a vCalAddress object.

    :returns str: Formatted list of attendees.
    """
    if isinstance(attendees, str):
        attendees = [attendees]
    return "\n".join(s.rjust(len(s) + 5) for s in map(_format_name, attendees))


def view(event):
    """Make a human readable summary of an iCalendar file.

    :returns str: Human readable summary.
    """
    summary = event.get("summary", default="")
    organizer = _format_name(event.get("organizer", default=""))
    attendees = _format_attendees(event.get("attendee", default=[]))
    location = event.get("location", default="")
    comment = event.get("comment", "")
    description = event.get("description", "").split("\n")
    description = "\n".join(s.rjust(len(s) + 5) for s in description)

    start = event.decoded("dtstart")
    if "duration" in event:
        end = event.decoded("dtend", default=start + event.decoded("duration"))
    else:
        end = event.decoded("dtend", default=start)
    duration = event.decoded("duration", default=end - start)
    if isinstance(start, datetime):
        start = start.astimezone()
    start = start.strftime("%c")
    if isinstance(end, datetime):
        end = end.astimezone()
    end = end.strftime("%c")

    return f"""    Organizer: {organizer}
    Attendees:
{attendees}
    Summary    : {summary}
    Starts     : {start}
    End        : {end}
    Duration   : {duration}
    Location   : {location}
    Comment    : {comment}
    Description:
{description}"""


def main():
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument(
        "calendar_files", nargs="+",
        help="one or more .ics files (use '-' for stdin)"
    )

    parser.add_argument(
        "--output",
        "-o",
        default="-",
        help="output file path (use '-' for stdout)"
    )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"{parser.prog} version {__version__}",
    )

    argv = parser.parse_args()

    # Open output file
    if argv.output == "-":
        output_file = sys.stdout
        close_output = False
    else:
        output_file = open(argv.output, "w", encoding="utf-8")
        close_output = True

    try:
        # Iterate over input paths
        for path in argv.calendar_files:
            if path == "-":
                f = sys.stdin
                close_input = False
            else:
                f = open(path, "r", encoding="utf-8-sig")
                close_input = True

            try:
                calendar = Calendar.from_ical(f.read())
                for event in calendar.walk("vevent"):
                    output_file.write(view(event) + "\n\n")
            finally:
                if close_input:
                    f.close()
    finally:
        if close_output:
            output_file.close()

__all__ = ["main", "view"]

if __name__ == "__main__":
    main()



"""This contains the IMAGE property from :rfc:`7986`."""

from __future__ import annotations

import base64

from icalendar.prop import vBinary, vText, vUri


class Image:
    """An image as URI or BINARY according to :rfc:`7986`.

    Value Type:
        URI or BINARY -- no default.  The value MUST be data
        with a media type of "image" or refer to such data.

    Description:
        This property specifies an image for an iCalendar
        object or a calendar component via a URI or directly with inline
        data that can be used by calendar user agents when presenting the
        calendar data to a user.  Multiple properties MAY be used to
        specify alternative sets of images with, for example, varying
        media subtypes, resolutions, or sizes.  When multiple properties
        are present, calendar user agents SHOULD display only one of them,
        picking one that provides the most appropriate image quality, or
        display none.  The "DISPLAY" parameter is used to indicate the
        intended display mode for the image.  The "ALTREP" parameter,
        defined in :rfc:`5545`, can be used to provide a "clickable" image
        where the URI in the parameter value can be "launched" by a click
        on the image in the calendar user agent.

    Arguments:
        uri: The URI of the image.
        b64data: The data of the image, base64 encoded.
        fmttype: The format type, e.g. ``"image/png"``.
        altrep: Link target of the image.
        display: The display mode, e.g. ``"BADGE"``.

    """

    @classmethod
    def from_property_value(cls, value: vUri | vBinary | vText):
        """Create an Image from a property value."""
        params: dict[str, str] = {}
        if not hasattr(value, "params"):
            raise TypeError("Value must be URI or BINARY.")
        value_type = value.params.get("VALUE", "").upper()
        if value_type == "URI" or isinstance(value, vUri):
            params["uri"] = str(value)
        elif isinstance(value, vBinary):
            params["b64data"] = value.obj
        elif value_type == "BINARY":
            params["b64data"] = str(value)
        else:
            raise TypeError(
                f"The VALUE parameter must be URI or BINARY, not {value_type!r}."
            )
        params.update(
            {
                "fmttype": value.params.get("FMTTYPE"),
                "altrep": value.params.get("ALTREP"),
                "display": value.params.get("DISPLAY"),
            }
        )
        return cls(**params)

    def __init__(
        self,
        uri: str | None = None,
        b64data: str | None = None,
        fmttype: str | None = None,
        altrep: str | None = None,
        display: str | None = None,
    ):
        """Create a new image according to :rfc:`7986`."""
        if uri is not None and b64data is not None:
            raise ValueError("Image cannot have both URI and binary data (RFC 7986)")
        if uri is None and b64data is None:
            raise ValueError("Image must have either URI or binary data")
        self.uri = uri
        self.b64data = b64data
        self.fmttype = fmttype
        self.altrep = altrep
        self.display = display

    @property
    def data(self) -> bytes | None:
        """Return the binary data, if available."""
        if self.b64data is None:
            return None
        return base64.b64decode(self.b64data)


__all__ = ["Image"]



"""Version information on icalendar.

This module provides a stable interface to the generated :file:`_version.py` file.

Version as a string:

.. code-block:: pycon

    >>> from icalendar import version
    >>> version  # doctest: +SKIP
    '7.0.0'

Version as a tuple:

.. code-block:: pycon

    >>> from icalendar import version_tuple
    >>> version_tuple  # doctest: +SKIP
    (7, 0, 0)

"""

try:
    from ._version import __version__, __version_tuple__, version, version_tuple
except ModuleNotFoundError:
    __version__ = version = "0.0.0dev0"
    __version_tuple__ = version_tuple = (0, 0, 0, "dev0")

__all__ = [
    "__version__",
    "__version_tuple__",
    "version",
    "version_tuple",
]



"""Tests for Issue 355.

see https://github.com/collective/icalendar/issues/355
"""

import pytest

from icalendar.parser import unescape_backslash


def test_facebook_link_is_correctly_parsed(events):
    """The facebook link must not be damaged.

    see https://github.com/collective/icalendar/pull/356#issuecomment-1222626128
    """
    assert (
        events.issue_355_url_escaping["DESCRIPTION"]
        == "https://www.facebook.com/events/756119502186737/?acontext=%7B%22source%22%3A5%2C%22action_history%22%3A[%7B%22surface%22%3A%22page%22%2C%22mechanism%22%3A%22main_list%22%2C%22extra_data%22%3A%22%5C%22[]%5C%22%22%7D]%2C%22has_source%22%3Atrue%7D"
    )


def test_other_facebook_link_is_correctly_parsed(events):
    """The facebook link must not be damaged.

    see https://github.com/collective/icalendar/pull/356#issuecomment-1265872696
    """
    expected_result = "https://www.facebook.com/events/756119502186737/?acontext=%7B%22source%22%3A5%2C%22action_history%22%3A[%7B%22surface%22%3A%22page%22%2C%22mechanism%22%3A%22main_list%22%2C%22extra_data%22%3A%22%5C%22[]%5C%22%22%7D]%2C%22has_source%22%3Atrue%7D"
    assert events.issue_355_url_escaping_2["DESCRIPTION"] == expected_result


def test_empty_quotes(events):
    """Make sure that empty quoted parameter values are supported."""
    assert events.issue_355_url_escaping_empty_param["ORGANIZER"].params["CN"] == ""


@pytest.mark.parametrize(
    ("input_string", "expected_result", "message"),
    [
        ("test 123", "test 123", "No escapes, should remain unchanged."),
        (r"\\n", "\\n", "No escape after escaped backslash."),
        (r"\\,", "\\,", "No escape after escaped backslash."),
        (r"\\;", "\\;", "No escape after escaped backslash."),
        (r"\\:", "\\:", "No escape after escaped backslash."),
        (r"\\N", "\\N", "No escape after escaped backslash."),
        ("-\\n-", "-\n-", "Newline escape."),
        ("-\\,", "-,", "Comma escape."),
        ("-\\;", "-;", "Semicolon escape."),
        ("-\\:", "-:", "Colon escape."),
        ("-\\N-", "-\n-", "Newline escape."),
    ],
)
def test_unescape_backslash(input_string, expected_result, message):
    """Test unescape_backslash function with various inputs."""
    assert unescape_backslash(input_string) == expected_result, (
        f"{message}: {input_string} -> {expected_result}"
    )



"""Use pytz timezones."""

from __future__ import annotations

from datetime import datetime, tzinfo
from typing import TYPE_CHECKING, Optional

import pytz
from pytz.tzinfo import DstTzInfo

from .provider import TZProvider

if TYPE_CHECKING:
    from dateutil.rrule import rrule

    from icalendar import prop
    from icalendar.cal import Timezone


class PYTZ(TZProvider):
    """Provide icalendar with timezones from pytz."""

    name = "pytz"

    def localize_utc(self, dt: datetime) -> datetime:
        """Return the datetime in UTC."""
        if getattr(dt, "tzinfo", False) and dt.tzinfo is not None:
            return dt.astimezone(pytz.utc)
        # assume UTC for naive datetime instances
        return pytz.utc.localize(dt)

    def localize(self, dt: datetime, tz: tzinfo) -> datetime:
        """Localize a datetime to a timezone."""
        return tz.localize(dt)

    def knows_timezone_id(self, id: str) -> bool:
        """Whether the timezone is already cached by the implementation."""
        return id in pytz.all_timezones

    def fix_rrule_until(self, rrule: rrule, ical_rrule: prop.vRecur) -> None:
        """Make sure the until value works for the rrule generated from the ical_rrule."""
        if not {"UNTIL", "COUNT"}.intersection(ical_rrule.keys()):
            # pytz.timezones don't know any transition dates after 2038
            # either
            rrule._until = datetime(2038, 12, 31, tzinfo=pytz.UTC)

    def create_timezone(self, tz: Timezone.Timezone) -> tzinfo:
        """Create a pytz timezone from the given information."""
        transition_times, transition_info = tz.get_transitions()
        name = tz.tz_name
        cls = type(
            name,
            (DstTzInfo,),
            {
                "zone": name,
                "_utc_transition_times": transition_times,
                "_transition_info": transition_info,
            },
        )
        return cls()

    def timezone(self, name: str) -> Optional[tzinfo]:
        """Return a timezone with a name or None if we cannot find it."""
        try:
            return pytz.timezone(name)
        except pytz.UnknownTimeZoneError:
            pass

    def uses_pytz(self) -> bool:
        """Whether we use pytz."""
        return True

    def uses_zoneinfo(self) -> bool:
        """Whether we use zoneinfo."""
        return False


__all__ = ["PYTZ"]



"""This tests the compatibility with RFC 7529.

CATEGORIES property

See
- https://github.com/collective/icalendar/issues/655
- https://www.rfc-editor.org/rfc/rfc7529.html
"""

from typing import Union

import pytest

from icalendar.cal.calendar import Calendar
from icalendar.cal.event import Event
from icalendar.cal.journal import Journal
from icalendar.cal.todo import Todo

CTJE = Union[Calendar, Event, Journal, Todo]


@pytest.fixture(params=[Event, Calendar, Todo, Journal])
def component(request):
    """An empty component with possible categories."""
    return request.param()


def test_no_categories_at_creation(component: CTJE):
    """An empty component has no categories."""
    assert "CATEGORIES" not in component
    assert component.categories == []


def test_add_one_category(component: CTJE):
    """Add one category."""
    component.add("categories", "Lecture")
    assert component.categories == ["Lecture"]


def test_add_multiple_categories(component: CTJE):
    """Add categories."""
    component.add("categories", ["Lecture", "Workshop"])
    assert component.categories == ["Lecture", "Workshop"]


def test_set_categories(component: CTJE):
    """Set categories."""
    component.categories = ["Lecture", "Workshop"]
    assert component.categories == ["Lecture", "Workshop"]


def test_modify_list(component: CTJE):
    """Modify the list and it still works."""
    component.categories = categories = ["cat1"]
    categories.append("cat2")
    assert component.categories == ["cat1", "cat2"]


def test_delete_categories(component: CTJE):
    """Delete categories."""
    component.categories = ["Lecture", "Workshop"]
    del component.categories
    assert "CATEGORIES" not in component
    assert component.categories == []


def test_manage_add_append_remove_categories(component: CTJE):
    """Manage multiple categories by merging them, then append
    and remove a category from the resulting set."""
    component.add("categories", ["c1", "c2"])
    component.add("categories", ["c3", "c4"])
    assert component.categories == ["c1", "c2", "c3", "c4"]
    component.categories.append("c5")
    assert component.categories == ["c1", "c2", "c3", "c4", "c5"]
    component.categories.remove("c2")
    assert component.categories == ["c1", "c3", "c4", "c5"]
    assert "c1,c3,c4,c5" in component.to_ical().decode()



from datetime import date, datetime

import pytest

from icalendar.cal.event import Event


def test_recurrence_properly_parsed(events):
    assert events.event_with_recurrence["rrule"] == {"COUNT": [100], "FREQ": ["DAILY"]}


@pytest.mark.parametrize(
    ("i", "exception_date"),
    [
        (0, datetime(1996, 4, 2, 1, 0)),
        (1, datetime(1996, 4, 3, 1, 0)),
        (2, datetime(1996, 4, 4, 1, 0)),
    ],
)
def test_exdate_properly_parsed(events, i, exception_date, in_timezone):
    assert events.event_with_recurrence["exdate"].dts[i].dt == in_timezone(
        exception_date, "UTC"
    )


def test_exdate_properly_marshalled(events):
    actual = events.event_with_recurrence["exdate"].to_ical()
    assert actual == b"19960402T010000Z,19960403T010000Z,19960404T010000Z"


# TODO: DOCUMENT BETTER!
# In this case we have multiple EXDATE definitions, one per line.
# Icalendar makes a list out of this instead of zipping it into one
# vDDDLists object. Actually, this feels correct for me, as it also
# allows to define different timezones per exdate line - but client
# code has to handle this as list and not blindly expecting to be able
# to call event['EXDATE'].to_ical() on it:
def test_exdate_formed_from_exdates_on_multiple_lines_is_a_list(events):
    exdate = events.event_with_recurrence_exdates_on_different_lines["exdate"]
    assert isinstance(exdate, list)


@pytest.mark.parametrize(
    ("i", "exception_date", "exception_date_ics"),
    [
        (0, datetime(2012, 5, 29, 10, 0), b"20120529T100000"),
        (1, datetime(2012, 4, 3, 10, 0), b"20120403T100000"),
        (2, datetime(2012, 4, 10, 10, 0), b"20120410T100000"),
        (3, datetime(2012, 5, 1, 10, 0), b"20120501T100000"),
        (4, datetime(2012, 4, 17, 10, 0), b"20120417T100000"),
    ],
)
def test_list_exdate_to_ical_is_inverse_of_from_ical(
    events, i, exception_date, exception_date_ics, in_timezone
):
    exdate = events.event_with_recurrence_exdates_on_different_lines["exdate"]
    assert exdate[i].dts[0].dt == in_timezone(exception_date, "Europe/Vienna")
    assert exdate[i].to_ical() == exception_date_ics


@pytest.mark.parametrize(
    ("freq", "byday", "dtstart", "expected"),
    [
        # Test some YEARLY BYDAY repeats
        (
            "YEARLY",
            "1SU",
            date(2016, 1, 3),  # 1st Sunday in year
            b"BEGIN:VEVENT\r\nSUMMARY:Event YEARLY 1SU\r\nDTSTART;VALUE=DATE:20160103\r\nRRULE:FREQ=YEARLY;BYDAY=1SU\r\nEND:VEVENT\r\n",
        ),
        (
            "YEARLY",
            "53MO",
            date(1984, 12, 31),  # 53rd Monday in (leap) year
            b"BEGIN:VEVENT\r\nSUMMARY:Event YEARLY 53MO\r\nDTSTART;VALUE=DATE:19841231\r\nRRULE:FREQ=YEARLY;BYDAY=53MO\r\nEND:VEVENT\r\n",
        ),
        (
            "YEARLY",
            "-1TU",
            date(1999, 12, 28),  # Last Tuesday in year
            b"BEGIN:VEVENT\r\nSUMMARY:Event YEARLY -1TU\r\nDTSTART;VALUE=DATE:19991228\r\nRRULE:FREQ=YEARLY;BYDAY=-1TU\r\nEND:VEVENT\r\n",
        ),
        (
            "YEARLY",
            "-17WE",
            date(2000, 9, 6),  # 17th-to-last Wednesday in year
            b"BEGIN:VEVENT\r\nSUMMARY:Event YEARLY -17WE\r\nDTSTART;VALUE=DATE:20000906\r\nRRULE:FREQ=YEARLY;BYDAY=-17WE\r\nEND:VEVENT\r\n",
        ),
        # Test some MONTHLY BYDAY repeats
        (
            "MONTHLY",
            "2TH",
            date(2003, 4, 10),  # 2nd Thursday in month
            b"BEGIN:VEVENT\r\nSUMMARY:Event MONTHLY 2TH\r\nDTSTART;VALUE=DATE:20030410\r\nRRULE:FREQ=MONTHLY;BYDAY=2TH\r\nEND:VEVENT\r\n",
        ),
        (
            "MONTHLY",
            "-3FR",
            date(2017, 5, 12),  # 3rd-to-last Friday in month
            b"BEGIN:VEVENT\r\nSUMMARY:Event MONTHLY -3FR\r\nDTSTART;VALUE=DATE:20170512\r\nRRULE:FREQ=MONTHLY;BYDAY=-3FR\r\nEND:VEVENT\r\n",
        ),
        (
            "MONTHLY",
            "-5SA",
            date(2053, 11, 1),  # 5th-to-last Saturday in month
            b"BEGIN:VEVENT\r\nSUMMARY:Event MONTHLY -5SA\r\nDTSTART;VALUE=DATE:20531101\r\nRRULE:FREQ=MONTHLY;BYDAY=-5SA\r\nEND:VEVENT\r\n",
        ),
        # Specifically test examples from the report of Issue #518
        # https://github.com/collective/icalendar/issues/518
        (
            "YEARLY",
            "9MO",
            date(2023, 2, 27),  # 9th Monday in year
            b"BEGIN:VEVENT\r\nSUMMARY:Event YEARLY 9MO\r\nDTSTART;VALUE=DATE:20230227\r\nRRULE:FREQ=YEARLY;BYDAY=9MO\r\nEND:VEVENT\r\n",
        ),
        (
            "YEARLY",
            "10MO",
            date(2023, 3, 6),  # 10th Monday in year
            b"BEGIN:VEVENT\r\nSUMMARY:Event YEARLY 10MO\r\nDTSTART;VALUE=DATE:20230306\r\nRRULE:FREQ=YEARLY;BYDAY=10MO\r\nEND:VEVENT\r\n",
        ),
    ],
)
def test_byday_to_ical(freq, byday, dtstart, expected):
    """Test the BYDAY rule is correctly processed by to_ical()."""
    event = Event()
    event.add("SUMMARY", f"Event {freq} {byday}")
    event.add("DTSTART", dtstart)
    event.add("RRULE", {"FREQ": [freq], "BYDAY": byday})
    assert event.to_ical() == expected



"""Tests for issue #860: Improved setters for start/duration/end.

This test suite defines the desired behavior for the improved setter implementations
as described in https://github.com/collective/icalendar/issues/860

The tests ensure RFC 5545 compliance, proper property locking mechanisms,
and backward compatibility with existing code.
"""

from datetime import date, datetime, timedelta

import pytest

from icalendar import Event, Todo
from icalendar.error import IncompleteComponent, InvalidCalendar


class TestDurationSetter:
    """Test the new duration setter functionality."""

    def test_event_duration_setter_with_start_locked(self):
        """Test setting duration with start locked (default behavior)."""
        event = Event()
        event.add("UID", "test-duration-setter")
        event.start = datetime(2026, 1, 1, 12, 0)
        event.end = datetime(2026, 1, 1, 14, 0)  # 2 hours initially

        # Setting duration should keep start locked and adjust end
        event.duration = timedelta(hours=3)

        assert event.start == datetime(2026, 1, 1, 12, 0)  # start unchanged
        assert event.end == datetime(2026, 1, 1, 15, 0)  # end adjusted
        assert event.duration == timedelta(hours=3)
        assert "DURATION" in event  # DURATION property should be set
        assert "DTEND" not in event  # DTEND should be removed

    def test_event_duration_setter_creates_duration_property(self):
        """Test that setting duration creates DURATION property and removes DTEND."""
        event = Event()
        event.add("UID", "test-duration-property")
        event.start = datetime(2026, 1, 1, 12, 0)
        event.end = datetime(2026, 1, 1, 14, 0)

        # Initially has DTEND
        assert "DTEND" in event
        assert "DURATION" not in event

        # Setting duration should switch to DURATION property
        event.duration = timedelta(hours=1)

        assert "DURATION" in event
        assert "DTEND" not in event
        assert event.duration == timedelta(hours=1)

    def test_todo_duration_setter_with_start_locked(self):
        """Test setting duration for Todo with start locked."""
        todo = Todo()
        todo.add("UID", "test-todo-duration")
        todo.start = datetime(2026, 1, 1, 12, 0)
        todo.end = datetime(2026, 1, 1, 14, 0)

        # Setting duration should keep start locked and adjust end
        todo.duration = timedelta(hours=4)

        assert todo.start == datetime(2026, 1, 1, 12, 0)  # start unchanged
        assert todo.end == datetime(2026, 1, 1, 16, 0)  # end adjusted
        assert todo.duration == timedelta(hours=4)
        assert "DURATION" in todo
        assert "DUE" not in todo

    def test_duration_setter_without_start_raises_error(self):
        """Test that setting duration without start raises appropriate error."""
        event = Event()
        event.add("UID", "test-no-start")

        with pytest.raises(IncompleteComponent):
            event.duration = timedelta(hours=2)


class TestExplicitLockingMethods:
    """Test the explicit locking methods set_start(), set_end(), set_duration()."""

    def test_set_duration_with_start_locked(self):
        """Test set_duration() with start locked (default)."""
        event = Event()
        event.add("UID", "test-set-duration")
        event.start = datetime(2026, 1, 1, 12, 0)
        event.end = datetime(2026, 1, 1, 14, 0)

        # Default behavior: lock start
        event.set_duration(timedelta(hours=3))

        assert event.start == datetime(2026, 1, 1, 12, 0)
        assert event.end == datetime(2026, 1, 1, 15, 0)
        assert event.duration == timedelta(hours=3)

    def test_set_duration_with_end_locked(self):
        """Test set_duration() with end locked."""
        event = Event()
        event.add("UID", "test-set-duration-end-locked")
        event.start = datetime(2026, 1, 1, 12, 0)
        event.end = datetime(2026, 1, 1, 14, 0)

        # Lock end, adjust start
        event.set_duration(timedelta(hours=3), locked="end")

        assert event.start == datetime(2026, 1, 1, 11, 0)  # start adjusted
        assert event.end == datetime(2026, 1, 1, 14, 0)  # end unchanged
        assert event.duration == timedelta(hours=3)

    def test_set_duration_convert_to_duration_property(self):
        """Test set_duration() with None to convert to DURATION property."""
        event = Event()
        event.add("UID", "test-convert-duration")
        event.start = datetime(2026, 1, 1, 12, 0)
        event.end = datetime(2026, 1, 1, 14, 0)

        # Convert to DURATION property
        event.set_duration(None)

        assert "DURATION" in event
        assert "DTEND" not in event
        assert event.duration == timedelta(hours=2)

    def test_set_start_with_duration_locked(self):
        """Test set_start() with duration locked."""
        event = Event()
        event.add("UID", "test-set-start")
        event.start = datetime(2026, 1, 1, 12, 0)
        event.add("DURATION", timedelta(hours=2))

        # Default behavior when DURATION exists: lock duration
        event.set_start(datetime(2026, 1, 1, 10, 0))

        assert event.start == datetime(2026, 1, 1, 10, 0)
        assert event.end == datetime(2026, 1, 1, 12, 0)  # end adjusted
        assert event.duration == timedelta(hours=2)  # duration unchanged

    def test_set_start_with_end_locked(self):
        """Test set_start() with end locked."""
        event = Event()
        event.add("UID", "test-set-start-end-locked")
        event.start = datetime(2026, 1, 1, 12, 0)
        event.end = datetime(2026, 1, 1, 14, 0)

        # Lock end, adjust duration
        event.set_start(datetime(2026, 1, 1, 10, 0), locked="end")

        assert event.start == datetime(2026, 1, 1, 10, 0)
        assert event.end == datetime(2026, 1, 1, 14, 0)  # end unchanged
        assert event.duration == timedelta(hours=4)  # duration adjusted

    def test_set_end_with_start_locked(self):
        """Test set_end() with start locked (default)."""
        event = Event()
        event.add("UID", "test-set-end")
        event.start = datetime(2026, 1, 1, 12, 0)
        event.end = datetime(2026, 1, 1, 14, 0)

        # Default behavior: lock start
        event.set_end(datetime(2026, 1, 1, 16, 0))

        assert event.start == datetime(2026, 1, 1, 12, 0)  # start unchanged
        assert event.end == datetime(2026, 1, 1, 16, 0)
        assert event.duration == timedelta(hours=4)  # duration adjusted

    def test_set_end_with_duration_locked(self):
        """Test set_end() with duration locked."""
        event = Event()
        event.add("UID", "test-set-end-duration-locked")
        event.start = datetime(2026, 1, 1, 12, 0)
        event.add("DURATION", timedelta(hours=2))

        # Lock duration, adjust start
        event.set_end(datetime(2026, 1, 1, 16, 0), locked="duration")

        assert event.start == datetime(2026, 1, 1, 14, 0)  # start adjusted
        assert event.end == datetime(2026, 1, 1, 16, 0)
        assert event.duration == timedelta(hours=2)  # duration unchanged


class TestImprovedPropertySetters:
    """Test the improved property setters with smart defaults."""

    def test_start_setter_with_duration_property(self):
        """Test start setter when DURATION property exists (should lock duration)."""
        event = Event()
        event.add("UID", "test-start-duration")
        event.start = datetime(2026, 1, 1, 12, 0)
        event.add("DURATION", timedelta(hours=2))

        # Smart default: lock duration when DURATION property exists
        event.start = datetime(2026, 1, 1, 10, 0)

        assert event.start == datetime(2026, 1, 1, 10, 0)
        assert event.end == datetime(2026, 1, 1, 12, 0)
        assert event.duration == timedelta(hours=2)

    def test_start_setter_with_dtend_property(self):
        """Test start setter when DTEND property exists (should lock end)."""
        event = Event()
        event.add("UID", "test-start-dtend")
        event.start = datetime(2026, 1, 1, 12, 0)
        event.end = datetime(2026, 1, 1, 14, 0)

        # Smart default: lock end when DTEND property exists
        event.start = datetime(2026, 1, 1, 10, 0)

        assert event.start == datetime(2026, 1, 1, 10, 0)
        assert event.end == datetime(2026, 1, 1, 14, 0)
        assert event.duration == timedelta(hours=4)

    def test_end_setter_with_duration_property(self):
        """Test end setter when DURATION property exists (should remove DURATION)."""
        event = Event()
        event.add("UID", "test-end-duration")
        event.start = datetime(2026, 1, 1, 12, 0)
        event.add("DURATION", timedelta(hours=2))

        # Setting end should remove DURATION and set DTEND
        event.end = datetime(2026, 1, 1, 16, 0)

        assert event.start == datetime(2026, 1, 1, 12, 0)
        assert event.end == datetime(2026, 1, 1, 16, 0)
        assert event.duration == timedelta(hours=4)
        assert "DTEND" in event
        assert "DURATION" not in event


class TestRFCCompliance:
    """Test that the improved setters maintain RFC 5545 compliance."""

    def test_dtend_and_duration_mutual_exclusion(self):
        """Test that DTEND and DURATION are mutually exclusive."""
        event = Event()
        event.add("UID", "test-mutual-exclusion")
        event.start = datetime(2026, 1, 1, 12, 0)
        event.end = datetime(2026, 1, 1, 14, 0)

        # Setting duration should remove DTEND
        event.duration = timedelta(hours=3)
        assert "DURATION" in event
        assert "DTEND" not in event

        # Setting end should remove DURATION
        event.end = datetime(2026, 1, 1, 16, 0)
        assert "DTEND" in event
        assert "DURATION" not in event

    def test_date_duration_validation(self):
        """Test that date DTSTART with time-based DURATION raises error."""
        event = Event()
        event.add("UID", "test-date-duration")
        event.start = date(2026, 1, 1)  # DATE type

        # Should raise error for non-day/week duration
        with pytest.raises(InvalidCalendar):
            event.duration = timedelta(hours=2)

        # Should work for day/week duration
        event.duration = timedelta(days=1)
        assert event.duration == timedelta(days=1)

    def test_datetime_type_consistency(self):
        """Test that DTSTART and DTEND must have same type."""
        event = Event()
        event.add("UID", "test-type-consistency")
        event.start = datetime(2026, 1, 1, 12, 0)  # datetime type

        # Should raise error for date end with datetime start
        event.end = date(2026, 1, 1)
        with pytest.raises(InvalidCalendar):
            _ = event.end  # Error occurs when accessing the property


class TestBackwardCompatibility:
    """Test that the improved setters maintain backward compatibility."""

    def test_existing_simple_usage_still_works(self):
        """Test that existing simple usage patterns still work."""
        event = Event()
        event.add("UID", "test-backward-compat")

        # Simple usage should still work
        event.start = datetime(2026, 1, 1, 12, 0)
        event.end = datetime(2026, 1, 1, 14, 0)

        assert event.start == datetime(2026, 1, 1, 12, 0)
        assert event.end == datetime(2026, 1, 1, 14, 0)
        assert event.duration == timedelta(hours=2)

    def test_property_access_compatibility(self):
        """Test that property access remains compatible."""
        event = Event()
        event.add("UID", "test-property-access")
        event.start = datetime(2026, 1, 1, 12, 0)
        event.add("DURATION", timedelta(hours=2))

        # Property access should work as before
        assert event.DTSTART is not None
        assert event.DURATION is not None
        assert event.DTEND is None

        # Computed properties should work
        assert event.start == datetime(2026, 1, 1, 12, 0)
        assert event.end == datetime(2026, 1, 1, 14, 0)
        assert event.duration == timedelta(hours=2)


class TestErrorConditions:
    """Test error conditions and edge cases for improved coverage."""

    def test_set_duration_invalid_locked_value(self):
        """Test that invalid locked values raise appropriate errors."""
        event = Event()
        event.add("UID", "test-invalid-locked")
        event.start = datetime(2026, 1, 1, 12, 0)

        with pytest.raises(
            ValueError, match="locked must be 'start' or 'end', not 'invalid'"
        ):
            event.set_duration(timedelta(hours=2), locked="invalid")

    def test_set_start_invalid_locked_value(self):
        """Test that invalid locked values raise appropriate errors."""
        event = Event()
        event.add("UID", "test-invalid-locked")
        event.start = datetime(2026, 1, 1, 12, 0)

        with pytest.raises(
            ValueError, match="locked must be 'duration', 'end', or None, not 'invalid'"
        ):
            event.set_start(datetime(2026, 1, 1, 14, 0), locked="invalid")

    def test_set_end_invalid_locked_value(self):
        """Test that invalid locked values raise appropriate errors."""
        event = Event()
        event.add("UID", "test-invalid-locked")
        event.start = datetime(2026, 1, 1, 12, 0)

        with pytest.raises(
            ValueError, match="locked must be 'start' or 'duration', not 'invalid'"
        ):
            event.set_end(datetime(2026, 1, 1, 14, 0), locked="invalid")

    def test_duration_setter_invalid_type(self):
        """Test that duration setter rejects invalid types."""
        event = Event()
        event.add("UID", "test-invalid-type")
        event.start = datetime(2026, 1, 1, 12, 0)

        with pytest.raises(TypeError, match="Use timedelta, not str"):
            event.duration = "2 hours"

    def test_set_duration_invalid_type(self):
        """Test that set_duration rejects invalid types."""
        event = Event()
        event.add("UID", "test-invalid-type")
        event.start = datetime(2026, 1, 1, 12, 0)

        with pytest.raises(TypeError, match="Use timedelta, not int"):
            event.set_duration(7200, locked="start")

    def test_todo_error_conditions(self):
        """Test error conditions for Todo class."""
        todo = Todo()
        todo.add("UID", "test-todo-errors")
        todo.start = datetime(2026, 1, 1, 12, 0)

        # Test invalid locked values
        with pytest.raises(
            ValueError, match="locked must be 'start' or 'end', not 'invalid'"
        ):
            todo.set_duration(timedelta(hours=2), locked="invalid")

        with pytest.raises(
            ValueError, match="locked must be 'duration', 'end', or None, not 'invalid'"
        ):
            todo.set_start(datetime(2026, 1, 1, 14, 0), locked="invalid")

        with pytest.raises(
            ValueError, match="locked must be 'start' or 'duration', not 'invalid'"
        ):
            todo.set_end(datetime(2026, 1, 1, 14, 0), locked="invalid")

    def test_set_start_auto_detect_no_existing_properties(self):
        """Test auto-detection when no existing properties exist."""
        event = Event()
        event.add("UID", "test-auto-detect")

        # When no DURATION or DTEND exists, should default to duration
        event.set_start(datetime(2026, 1, 1, 12, 0), locked=None)
        assert event.start == datetime(2026, 1, 1, 12, 0)

        # Adding end should trigger duration calculation
        event.end = datetime(2026, 1, 1, 14, 0)
        # The auto-detect should have preserved the end time calculation
